#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkx.h>
#include "guiutils.h"
#include "imgview.h"

#include "images/icon_zoom_in_16x16.xpm"
#include "images/icon_zoom_out_16x16.xpm"
#include "images/icon_zoom_onetoone_16x16.xpm"
#include "images/icon_zoom_tofit_16x16.xpm"

/*
#include "images/icon_zoom_in_20x20.xpm"
#include "images/icon_zoom_out_20x20.xpm"
 */
#include "images/icon_zoom_onetoone_20x20.xpm"
#include "images/icon_zoom_tofit_20x20.xpm"

#include "iv.xpm"


/* Image IO. */
static imgview_image_struct *ImgViewImageNew(
	gint width, gint height, gint depth
);
static void ImgViewImageClear(
        imgview_image_struct *img, guint32 v
);
static void ImgViewImageSendRectangle(
	imgview_image_struct *image, GdkDrawable *d, GdkGC *gc,
	GdkRectangle *rect
);
static void ImgViewImageSend(
	imgview_image_struct *image, GdkDrawable *d, GdkGC *gc
);
static void ImgViewImageDelete(imgview_image_struct *image);

/* Callbacks. */
static gint ImgViewCloseCB(GtkWidget *widget, GdkEvent *event, gpointer data);
static void ImgViewDestroyCB(GtkObject *object, gpointer data);
static void ImgViewZoomInPressedCB(GtkWidget *widget, gpointer data);
static gint ImgViewZoomInTOCB(gpointer data);
static void ImgViewZoomOutPressedCB(GtkWidget *widget, gpointer data);
static gint ImgViewZoomOutTOCB(gpointer data);
static void ImgViewZoomReleasedCB(GtkWidget *widget, gpointer data);
static void ImgViewZoomOneToOneCB(GtkWidget *widget, gpointer data);
static void ImgViewZoomToFitCB(GtkWidget *widget, gpointer data);
static void ImgViewToolBarToggleCB(GtkWidget *widget, gpointer data);
static void ImgViewValuesToggleCB(GtkWidget *widget, gpointer data);
static void ImgViewAdjustmentChangedCB(
        GtkAdjustment *adjustment, gpointer data
);
static gint ImgViewViewEventCB(
	GtkWidget *widget, GdkEvent *event, gpointer data
);
static gint ImgViewExposeCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
);

/* Utilities and operations. */
static void ImgViewRealizeChange(imgview_struct *iv);
static void ImgViewZoomItterate(imgview_struct *iv, gint dz);
static void ImgViewZoomRectangular(
	imgview_struct *iv,
	gint x0, gint x1,
	gint y0, gint y1
);

static void ImgViewBlitView(imgview_struct *iv);
static void ImgViewWMIconUpdate(imgview_struct *iv);
static void ImgViewBufferRecreate(imgview_struct *iv);

/* Public. */
gbool ImgViewToplevelIsWindow(imgview_struct *iv);
GtkWidget *ImgViewGetToplevelWidget(imgview_struct *iv);
GtkDrawingArea *ImgViewGetViewWidget(imgview_struct *iv);
GtkMenu *ImgViewGetMenuWidget(imgview_struct *iv);
void ImgViewClear(imgview_struct *iv);
gint ImgViewLoad(
	imgview_struct *iv,
	gint width, gint height,
	gint bytes_per_line,	/* Can be 0 to auto calculate. */
	gint format,		/* One of IMGVIEW_FORMAT_*. */
	const guint8 *data
);

imgview_struct *ImgViewNew(
	gbool show_toolbar,
	gbool show_values,
	gbool show_image_on_wm_icon,
	gbool toplevel_is_window,
	GtkWidget **toplevel_rtn
);
void ImgViewReset(imgview_struct *iv, gbool need_unmap);
void ImgViewUpdateMenus(imgview_struct *iv);
void ImgViewMap(imgview_struct *iv);
void ImgViewUnmap(imgview_struct *iv);
void ImgViewDelete(imgview_struct *iv);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)		(((x) < 0) ? ((x) * -1) : (x))

#ifndef GDK_BUTTON1
# define GDK_BUTTON1	1
#endif

#ifndef GDK_BUTTON2
# define GDK_BUTTON2	2
#endif

#ifndef GDK_BUTTON3
# define GDK_BUTTON3	3
#endif


/* Title. */
#define IMGVIEW_TITLE		"Image Viewer"

/* Minimum width. */
#define IMGVIEW_MIN_WIDTH	200
#define	IMGVIEW_MIN_HEIGHT	(IMGVIEW_MIN_WIDTH * 0.75)

/* WM icon size in pixels. */
#define IMGVIEW_WM_ICON_WIDTH	48
#define IMGVIEW_WM_ICON_HEIGHT	48

/* Zoom out maximum. */
#define IMGVIEW_ZOOM_MIN	0.01

/* Zoom in maximum. */
#define IMGVIEW_ZOOM_MAX	50.0

/* Zoom rate (in coeff per cycle). */
#define IMGVIEW_ZOOM_RATE	0.0025


/* Toolbar height in pixels. */
#define IMGVIEW_TOOLBAR_HEIGHT	(16 + (2 * 2))


/* Zoom repeat timeout interval in ms. */
#define IMGVIEW_ZOOM_TIMEOUT_INT		110
#define IMGVIEW_ZOOM_INITIAL_TIMEOUT_INT	390

/* Zoom itteration pixels for button or keypresses. */
#define IMGVIEW_ZOOM_ITTERATION_PIXELS		50


/*
 *	Creates a new image.
 */
static imgview_image_struct *ImgViewImageNew(
        gint width, gint height, gint depth
)
{
	gint bpl;

	imgview_image_struct *img = (imgview_image_struct *)calloc(
	    1, sizeof(imgview_image_struct)
	);
	if(img == NULL)
	    return(NULL);

	if((width < 1) || (height < 1))
	{
	    g_free(img);
	    return(NULL);
	}

	img->depth = 24;	/* Gdk's RGB uses 24 bit data types. */
	img->bpp = 4;		/* 4 bytes since using GDK 32 bits rgb. */

	/* Calculate bytes per line. */
	bpl = img->bpp * width;
	/* Align bytes per line to 4 bytes by incrementing it untill it
	 * matches the next 4 byte align value
	 */
	if(bpl > 4)
	    while((bpl % 4) != 0)
		bpl++;
	else
	    bpl = 4;
	img->bpl = bpl;

	/* Set width and height in units of pixels. */
	img->width = width;
	img->height = height;

	/* Allocate image buffer. */
	img->mem = (guint8 *)g_malloc(img->bpl * img->height);
	if(img->mem == NULL)
	{
	    ImgViewImageDelete(img);
	    return(NULL);
	}

	return(img);
}

/*
 *	Clears the given image with the given pixel value v.
 *
 *	The pixel value v must be 32 bits in size and in the format
 *	RGBA.
 */
static void ImgViewImageClear(
	imgview_image_struct *img, guint32 v
)
{
	gint x, y, bc;
	gint width, height;
	gint bpl, bpp, index_offset;
	guint8 *buf, *buf_ptr;
	guint32 vbuf = v;
	const guint8 *vptr;


	if(img == NULL)
	    return;

	buf = img->mem;
	if(buf == NULL)
	    return;

	width = img->width,
	height = img->height;
	bpp = img->bpp;
	bpl = img->bpl;

	/* Itterate through rows. */
	for(y = 0; y < height; y++)
	{
	    index_offset = y * bpl;	/* Cur line index offset in bytes. */
	    for(x = 0; x < width; x++)
	    {
		/* Get pointer to target buffer and start of given pixel
		 * value.
		 */
		buf_ptr = &buf[index_offset + (x * bpp)];
		vptr = (const guint8 *)&vbuf;

		for(bc = 0; bc < bpp; bc++)
		    *buf_ptr++ = *vptr++;
	    }
	}
}

/*
 *	Sends the given rectangular area of the given image to the
 *	drawable.
 */
static void ImgViewImageSendRectangle(
        imgview_image_struct *img, GdkDrawable *d, GdkGC *gc,
        GdkRectangle *rect
)
{
        guchar *rgb_buf;

        if((img == NULL) || (d == NULL) || (gc == NULL) ||
           (rect == NULL)
	)
            return;

        rgb_buf = img->mem;
        if(rgb_buf == NULL)
            return;

/* Rectangles don't work too well, need to be upper left oriented. */
	rect->width += rect->x;
	rect->height += rect->y;
	rect->x = 0;
	rect->y = 0;

        gdk_draw_rgb_32_image(
            d, gc,
	    rect->x, rect->y,
            rect->width, rect->height,
            GDK_RGB_DITHER_NORMAL,
            rgb_buf,
            img->bpl
        );
}

/*
 *	Sends the given image to the drawable.
 */
static void ImgViewImageSend(
        imgview_image_struct *img, GdkDrawable *d, GdkGC *gc
)
{
	GdkRectangle rect;

	if((img == NULL) || (d == NULL) || (gc == NULL))
	    return;

	rect.x = 0;
	rect.y = 0;
	rect.width = img->width;
	rect.height = img->height;

	ImgViewImageSendRectangle(img, d, gc, &rect);
}

/*
 *	Deallocates the given image and all its resources.
 */
static void ImgViewImageDelete(imgview_image_struct *img)
{
	if(img == NULL)
	    return;

	/* Deallocate image buffer. */
	if(img->mem != NULL)
	{
	    guint8 *tm = img->mem;
	    img->mem = NULL;
	    g_free(tm);
	}

	/* Deallocate structure itself. */
	g_free(img);
}


/*
 *	Close "delete_event" callback.
 */
static gint ImgViewCloseCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	imgview_struct *iv = (imgview_struct *)data;
	if(iv == NULL)
	    return(FALSE);

/* Let calling function determine what close does, do not unmap or
 * do anything here!
 */
/*	ImgViewUnmap(iv); */
	return(TRUE);
}

/*
 *	Destroy event callback.
 */
static void ImgViewDestroyCB(GtkObject *object, gpointer data)
{
	return;
}

/*
 *	Zoom in "pressed" callback.
 */
static void ImgViewZoomInPressedCB(GtkWidget *widget, gpointer data)
{
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

        /* Call timeout callback which will itterate once and set up
         * timeout callback to call itself again.
         */
	ImgViewZoomInTOCB(iv);
}

/*
 *	Zoom in timeout callback.
 */
static gint ImgViewZoomInTOCB(gpointer data)
{
	gbool is_initial = FALSE;
	imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return(FALSE);

        /* Zoom in. */
        ImgViewZoomItterate(iv, -IMGVIEW_ZOOM_ITTERATION_PIXELS);

        /* Clip bounds, update adjustments and redraw. */
        ImgViewRealizeChange(iv);
        ImgViewExposeCB(iv->view_da, NULL, iv);

        /* Flush output and reset timeout for next itteration. */
        gdk_flush();
        if(iv->view_zoom_toid != (guint)-1)
            gtk_timeout_remove(iv->view_zoom_toid);
	else
	    is_initial = TRUE;
        iv->view_zoom_toid = gtk_timeout_add(
            (guint32)(is_initial ?
                IMGVIEW_ZOOM_INITIAL_TIMEOUT_INT : IMGVIEW_ZOOM_TIMEOUT_INT
            ),
            (GtkFunction)ImgViewZoomInTOCB,
            iv
        );

	return(FALSE);
}

/*
 *	Zoom out "pressed" callback.
 */
static void ImgViewZoomOutPressedCB(GtkWidget *widget, gpointer data)
{
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

	/* Call timeout callback which will itterate once and set up
	 * timeout callback to call itself again.
	 */
	ImgViewZoomOutTOCB(iv);
}

/*
 *	Zoom out timeout callback.
 */
static gint ImgViewZoomOutTOCB(gpointer data)
{
	gbool is_initial = FALSE;
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return(FALSE);

	/* Zoom out. */
	ImgViewZoomItterate(iv, IMGVIEW_ZOOM_ITTERATION_PIXELS);

	/* Clip bounds, update adjustments and redraw. */
	ImgViewRealizeChange(iv);
	ImgViewExposeCB(iv->view_da, NULL, iv);

	/* Flush output and reset timeout for next itteration. */
	gdk_flush();
        if(iv->view_zoom_toid != (guint)-1)
	    gtk_timeout_remove(iv->view_zoom_toid);
	else
	    is_initial = TRUE;
	iv->view_zoom_toid = gtk_timeout_add(
	    (guint32)(is_initial ? 
		IMGVIEW_ZOOM_INITIAL_TIMEOUT_INT : IMGVIEW_ZOOM_TIMEOUT_INT
	    ),
	    (GtkFunction)ImgViewZoomOutTOCB,
	    iv
	);

        return(FALSE);
}

/*
 *	Zoom in or zoom out button "released" callback.
 */
static void ImgViewZoomReleasedCB(GtkWidget *widget, gpointer data)
{
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

        if(iv->view_zoom_toid != (guint)-1)
        {
            gtk_timeout_remove(iv->view_zoom_toid);
	    iv->view_zoom_toid = (guint)-1;
        }
}


/*
 *	Zoom one to one callback.
 */
static void ImgViewZoomOneToOneCB(GtkWidget *widget, gpointer data)
{
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

        if(iv->orig_img == NULL)
            return;

	/* Reset zoom to 1.0. */
	iv->view_zoom = 1.0;

        /* Clip bounds, update adjustments and redraw. */
        ImgViewRealizeChange(iv);
        ImgViewExposeCB(iv->view_da, NULL, iv);
}

/*
 *	Zoom to fit callback.
 */
static void ImgViewZoomToFitCB(GtkWidget *widget, gpointer data)
{
	GtkAdjustment *xadj, *yadj;
	imgview_image_struct *tar_img;
	gint src_width, src_height;
	gdouble zoom;
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

	if(iv->orig_img == NULL)
	    return;

        xadj = iv->view_x_adj;
        yadj = iv->view_y_adj;
        tar_img = iv->view_img;

	if((xadj == NULL) || (yadj == NULL) || (tar_img == NULL))
	    return;

	if((tar_img->width < 1) || (tar_img->height < 1))
	    return;

	src_width = (gint)(xadj->upper - xadj->lower);
	src_height = (gint)(yadj->upper - yadj->lower);
	if((src_width < 1) || (src_height < 1))
	    return;

	zoom = (gdouble)tar_img->width / (gdouble)src_width;
	if(zoom <= 0.0)
	    return;

	if((tar_img->height / zoom) < src_height)
	    zoom = (gdouble)tar_img->height / (gdouble)src_height;

	iv->view_zoom = zoom;

        /* Clip bounds, update adjustments and redraw. */
        ImgViewRealizeChange(iv);
        ImgViewExposeCB(iv->view_da, NULL, iv);

}

/*
 *      Show toolbar toggle callback.
 */
static void ImgViewToolBarToggleCB(GtkWidget *widget, gpointer data)
{
	static gbool reenterant = FALSE;
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

        ImgViewShowToolBar(iv, !iv->toolbar_map_state);

	reenterant = FALSE;
}

/*
 *      Show values toggle callback.
 */
static void ImgViewValuesToggleCB(GtkWidget *widget, gpointer data)
{
        static gbool reenterant = FALSE;
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

        if(reenterant)
            return;
        else
            reenterant = TRUE;

	iv->show_values = !iv->show_values;
	ImgViewRealizeChange(iv);	/* To update info_label. */
	ImgViewUpdateMenus(iv);
	ImgViewExposeCB(iv->view_da, NULL, iv);

        reenterant = FALSE;
}


/*
 *	Adjustment "changed" signal callback.
 */
static void ImgViewAdjustmentChangedCB(
	GtkAdjustment *adjustment, gpointer data
)
{
        imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

	/* Redraw view drawing area when an adjustment has changed. */
	ImgViewExposeCB(iv->view_da, NULL, iv);
}

/*
 *	View widget event handler (except for event "expose_event", see
 *	ImgViewExposeCB() for that handler).
 */
static gint ImgViewViewEventCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gint etype, status = FALSE;
	gbool need_grab_pointer = FALSE;
	GdkCursor *cur = NULL;
	imgview_image_struct *src_img, *tar_img;
	GdkWindow *window;
	GtkAdjustment *xadj, *yadj;
	GdkEventConfigure *configure;
	GdkEventKey *key;
	GdkEventButton *button;
	GdkEventMotion *motion;
	gbool keystate;
	guint keyval, motion_timmer;
	gint x, y;
        GdkModifierType mask;
        imgview_struct *iv = (imgview_struct *)data;
        if((widget == NULL) || (event == NULL) || (iv == NULL))
            return(status);


	/* Get view adjustments (may be NULL). */
	xadj = iv->view_x_adj;
	yadj = iv->view_y_adj;

	/* Get source and target images (may be NULL). */
	src_img = iv->orig_img;
	tar_img = iv->view_img;

	/* Get view widget GdkWindow. */
	if(GTK_WIDGET_NO_WINDOW(widget))
	    window = NULL;
	else
	    window = widget->window;

	/* Get event type. */
	etype = *(gint *)event;

	/* Handle by event type. */
	switch(etype)
	{
          case GDK_CONFIGURE:
            configure = (GdkEventConfigure *)event;
	    /* Destroy old view image buffer and create a new one based on
	     * the new size of the view_da.
	     */
            ImgViewBufferRecreate(iv);
	    /* Need to update scrollbar about the new view_da (target image)
	     * size.
	     */
	    ImgViewRealizeChange(iv);
            status = TRUE;
            break;

	  case GDK_ENTER_NOTIFY:
	    /* Set up widget to grab focus so key_press_event and
	     * and key_release_event events will be sent to it.
	     */
            gtk_widget_grab_focus(widget);
	    break;

	  case GDK_LEAVE_NOTIFY:
	    GTK_WIDGET_UNSET_FLAGS(widget, GTK_HAS_FOCUS);
	    break;

          case GDK_KEY_PRESS: case GDK_KEY_RELEASE:
            key = (GdkEventKey *)event;
	    keyval = key->keyval;
	    keystate = ((etype == GDK_KEY_PRESS) ? TRUE : FALSE);
	    /* Handle by keyval. */
	    /* ALT? */
            if((keyval == GDK_Alt_L) || (keyval == GDK_Alt_R))
            {
                if(keystate)
                    iv->modifiers |= IMGVIEW_MODIFIER_ALT;
                else
                    iv->modifiers &= ~IMGVIEW_MODIFIER_ALT;
                status = TRUE;
            }
	    /* CTRL? */
            else if((keyval == GDK_Control_L) || (keyval == GDK_Control_R))
            {
                if(keystate)
                    iv->modifiers |= IMGVIEW_MODIFIER_CTRL;
                else
                    iv->modifiers &= ~IMGVIEW_MODIFIER_CTRL;
                status = TRUE;
            }
	    /* SHIFT? */
            else if((keyval == GDK_Shift_L) || (keyval == GDK_Shift_R))
            {
                if(keystate)
                    iv->modifiers |= IMGVIEW_MODIFIER_SHIFT;
                else
                    iv->modifiers &= ~IMGVIEW_MODIFIER_SHIFT;
		status = TRUE;
	    }
	    /* Zoom in? */
	    else if((keyval == GDK_plus) || (keyval == GDK_equal) ||
                    (keyval == GDK_KP_Add)
	    )
	    {
                if(keystate)
		{
                    ImgViewZoomItterate(
			iv, -IMGVIEW_ZOOM_ITTERATION_PIXELS
		    );
                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
		}
                status = TRUE;
	    }
            /* Zoom in fast? */
            else if((keyval == GDK_Page_Up) ||
                    (keyval == GDK_KP_Page_Up)
            )
            {
                if(keystate)
                {
                    ImgViewZoomItterate(
                        iv, -IMGVIEW_ZOOM_ITTERATION_PIXELS * 4
                    );
                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
                }
                status = TRUE;
            }
            /* Zoom out? */
            else if((keyval == GDK_minus) || (keyval == GDK_underscore) ||
                    (keyval == GDK_KP_Subtract)
	    )
            {
                if(keystate)
                {
                    ImgViewZoomItterate(
			iv, IMGVIEW_ZOOM_ITTERATION_PIXELS
		    );
                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
                }
                status = TRUE;
            }
            /* Zoom out fast? */
            else if((keyval == GDK_Page_Down) ||
                    (keyval == GDK_KP_Page_Down)
	    )
            {
                if(keystate)
                {
                    ImgViewZoomItterate(
                        iv, IMGVIEW_ZOOM_ITTERATION_PIXELS * 4
                    );
                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
                }
                status = TRUE;
            }
	    /* Translate up. */
	    else if((keyval == GDK_Up) || (keyval == GDK_KP_Up))
	    {
		if(keystate && (xadj != NULL) && (yadj != NULL))
		{
		    gdouble zoom = iv->view_zoom;
		    const gdouble zoom_min = IMGVIEW_ZOOM_MIN;

                    /* Sanitize current zoom. */
                    if(zoom < zoom_min)
                        zoom = zoom_min;

                    /* Translate. */
                    yadj->value -= (gfloat)yadj->step_increment / zoom;

                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
		}
		/* Stop signal so GTK+ does not later handle it and change
		 * the focus widget, since this is also a focus change key.
		 */
		gtk_signal_emit_stop_by_name(
		    GTK_OBJECT(widget),
		    keystate ? "key_press_event" : "key_release_event"
		);
                status = TRUE;
	    }
            /* Translate down. */
            else if((keyval == GDK_Down) || (keyval == GDK_KP_Down))
            {
                if(keystate && (xadj != NULL) && (yadj != NULL))
                {
                    gdouble zoom = iv->view_zoom;
                    const gdouble zoom_min = IMGVIEW_ZOOM_MIN;

                    /* Sanitize current zoom. */
                    if(zoom < zoom_min)
                        zoom = zoom_min;

                    /* Translate. */
                    yadj->value += (gfloat)yadj->step_increment / zoom;

                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
                }
                /* Stop signal so GTK+ does not later handle it and change
                 * the focus widget, since this is also a focus change key.
                 */
                gtk_signal_emit_stop_by_name(
                    GTK_OBJECT(widget),
                    keystate ? "key_press_event" : "key_release_event"
                );
                status = TRUE;
            }
            /* Translate left. */
            else if((keyval == GDK_Left) || (keyval == GDK_KP_Left))
            {
                if(keystate && (xadj != NULL) && (yadj != NULL))
                {
                    gdouble zoom = iv->view_zoom;
                    const gdouble zoom_min = IMGVIEW_ZOOM_MIN;

                    /* Sanitize current zoom. */
                    if(zoom < zoom_min)
                        zoom = zoom_min;

                    /* Translate. */
                    xadj->value -= (gfloat)xadj->step_increment / zoom;

                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
                }
                /* Stop signal so GTK+ does not later handle it and change
                 * the focus widget, since this is also a focus change key.
                 */
                gtk_signal_emit_stop_by_name(
                    GTK_OBJECT(widget),
                    keystate ? "key_press_event" : "key_release_event"
                );
                status = TRUE;
            }
            /* Translate right. */
            else if((keyval == GDK_Right) || (keyval == GDK_KP_Right))
            {
                if(keystate && (xadj != NULL) && (yadj != NULL))
                {
                    gdouble zoom = iv->view_zoom;
                    const gdouble zoom_min = IMGVIEW_ZOOM_MIN;

                    /* Sanitize current zoom. */
                    if(zoom < zoom_min)
                        zoom = zoom_min;

                    /* Translate. */
                    xadj->value += (gfloat)xadj->step_increment / zoom;

                    ImgViewRealizeChange(iv);
                    ImgViewExposeCB(iv->view_da, NULL, iv);
                }
                /* Stop signal so GTK+ does not later handle it and change
                 * the focus widget, since this is also a focus change key.
                 */
                gtk_signal_emit_stop_by_name(
                    GTK_OBJECT(widget),
                    keystate ? "key_press_event" : "key_release_event"
                );
                status = TRUE;
            }
	    /* Zoom to fit? */
	    else if((keyval == GDK_End) || (keyval == GDK_KP_End))
	    {
		ImgViewZoomToFitCB(widget, iv);
		status = TRUE;
	    }
            /* Zoom one to one? */
            else if((keyval == GDK_Home) || (keyval == GDK_KP_Home))
            {
		ImgViewZoomOneToOneCB(widget, iv);
		status = TRUE;
            }
	    /* Event handled? */
	    if(status)
	    {
		if(keystate)
		{
		    if(iv->modifiers & IMGVIEW_MODIFIER_ALT)
			cur = iv->zoom_cur;
                    else if(iv->modifiers & IMGVIEW_MODIFIER_SHIFT)
                        cur = iv->zoom_rectangle_cur;
		    else if(iv->modifiers & IMGVIEW_MODIFIER_CTRL)
			cur = iv->translate_cur;
		}
                /* Update cursor. */
                if(window != NULL)
                    gdk_window_set_cursor(window, cur);
	    }
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    cur = NULL;
            switch(button->button)
            {
              case GDK_BUTTON1:
		if(iv->modifiers & IMGVIEW_MODIFIER_SHIFT)
		{
		    iv->drag_mode = IMGVIEW_DRAG_MODE_ZOOM_RECTANGLE;
		    cur = iv->zoom_rectangle_cur;
		}
		else if(iv->modifiers & IMGVIEW_MODIFIER_ALT)
		{
		    iv->drag_mode = IMGVIEW_DRAG_MODE_ZOOM;
                    cur = iv->zoom_cur;
		}
		else
		{
		    /* Modifier with or without CTRL will translate. */
		    iv->drag_mode = IMGVIEW_DRAG_MODE_TRANSLATE;
		    cur = iv->translate_cur;
		}
		need_grab_pointer = TRUE;
                status = TRUE;
                break;

              case GDK_BUTTON2:
                if(iv->modifiers & IMGVIEW_MODIFIER_CTRL)
                {
                    /* Zoom 1:1. */
                    ImgViewZoomOneToOneCB(widget, data);
                }
		else if(iv->modifiers & IMGVIEW_MODIFIER_SHIFT)
		{
		    /* Zoom to fit. */
		    ImgViewZoomToFitCB(widget, data);
		}
		else
		{
		    /* Drag zoom. */
		    iv->drag_mode = IMGVIEW_DRAG_MODE_ZOOM;
		    cur = iv->zoom_cur;
		}
		need_grab_pointer = TRUE;
                status = TRUE;
                break;

              case GDK_BUTTON3:
		if(iv->menu != NULL)
		{
		    GtkMenu *menu = GTK_MENU(iv->menu);
                    gtk_menu_popup(
                        menu,
                        NULL, NULL, NULL, NULL,
                        button->button, button->time
                    );
		}
                status = TRUE;
                break;
	    }
	    /* Set cursor. */
	    if(window != NULL)
		gdk_window_set_cursor(window, cur);

	    /* Reset last drag positions on any button press. */
	    iv->drag_last_x = button->x;
            iv->drag_last_y = button->y;

	    /* Record last drag rectangle position. */
	    iv->drag_zoom_rectangle_start_x =
		iv->drag_zoom_rectangle_cur_x = button->x;
	    iv->drag_zoom_rectangle_start_y =
		iv->drag_zoom_rectangle_cur_y = button->y;

	    /* Reset last motion event handled time. */
            iv->last_motion_time = 0;

	    /* Grab pointer? */
	    if(need_grab_pointer && (window != NULL))
	    {
		gdk_flush();
                gdk_pointer_grab(
                    window,
                    FALSE,
                    GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
                        GDK_POINTER_MOTION_MASK,
                    /* w->window, */ NULL,
                    GDK_NONE,
                    button->time
                );
		gdk_flush();
	    }
            break;

          case GDK_BUTTON_RELEASE:
            button = (GdkEventButton *)event;

            /* Ungrab pointer. */
            gdk_flush();
            gdk_pointer_ungrab(button->time);
            gdk_flush();
            while(gdk_pointer_is_grabbed())
                { gdk_pointer_ungrab(GDK_CURRENT_TIME); gdk_flush(); }

	    /* Post drag handling. */
	    switch(iv->drag_mode)
	    {
	      case IMGVIEW_DRAG_MODE_ZOOM:
                /* Need to reset drag mode before redraw. */
                iv->drag_mode = IMGVIEW_DRAG_MODE_NONE;

                /* Redraw. */
                ImgViewExposeCB(iv->view_da, NULL, iv);
		break;

              case IMGVIEW_DRAG_MODE_ZOOM_RECTANGLE:
		/* Zoom to rectangle. */
		ImgViewZoomRectangular(
		    iv,
		    button->x, iv->drag_zoom_rectangle_start_x,
		    button->y, iv->drag_zoom_rectangle_start_y
		);

                /* Need to reset drag mode before redraw. */
                iv->drag_mode = IMGVIEW_DRAG_MODE_NONE;

		/* Clip bounds, update adjustments and redraw. */
		ImgViewRealizeChange(iv);
		ImgViewExposeCB(iv->view_da, NULL, iv);
		break;
	    }

	    /* Cancel any drag modes whenever a button is released. */
	    iv->drag_mode = IMGVIEW_DRAG_MODE_NONE;

            /* Reset last drag positions on any button release. */
            iv->drag_last_x = button->x;
            iv->drag_last_y = button->y;

            /* Reset last drag rectangle position. */
            iv->drag_zoom_rectangle_start_x =
                iv->drag_zoom_rectangle_cur_x = button->x;
            iv->drag_zoom_rectangle_start_y =
                iv->drag_zoom_rectangle_cur_y = button->y;

            /* Reset last motion event handled time. */
            iv->last_motion_time = 0;

            /* Reset cursor only if no key modifiers are held. */
            if((iv->modifiers == 0) && (window != NULL))
                gdk_window_set_cursor(window, NULL);

            status = TRUE;
            break;

          case GDK_MOTION_NOTIFY:
            motion = (GdkEventMotion *)event;
            if(motion->is_hint)
            {
                if(iv->drag_mode != IMGVIEW_DRAG_MODE_NONE)
                    gdk_window_get_pointer(
                        motion->window, &x, &y, &mask
                    );
            }
            else
            {
                x = motion->x;
                y = motion->y;
                mask = motion->state;
            }

/*
printf("Motion last:%i ev:%i ms\n",
 iv->last_motion_time, motion->time
);
 */
	    /* Too soon to process motion event? */
	    if(iv->last_motion_time > 0)
	    {
/*
		if(iv->last_motion_time > motion->time)
		    break;
 */
	    }

	    /* Initialize motion timmer. */
	    motion_timmer = gdk_time_get();

	    /* Handle by drag mode. */
	    switch(iv->drag_mode)
	    {
              case IMGVIEW_DRAG_MODE_TRANSLATE:
		if((xadj != NULL) && (yadj != NULL))
		{
		    gint	dx = iv->drag_last_x - x,
				dy = iv->drag_last_y - y;
		    gdouble zoom = iv->view_zoom;
		    const gdouble zoom_min = IMGVIEW_ZOOM_MIN;


		    /* Sanitize current zoom. */
		    if(zoom < zoom_min)
			zoom = zoom_min;

		    /* Translate. */
		    xadj->value += (gfloat)dx / zoom;
                    yadj->value += (gfloat)dy / zoom;

                    /* Clip bounds and update adjustments. */
		    ImgViewRealizeChange(iv);
		    ImgViewExposeCB(iv->view_da, NULL, iv);
		}
		status = TRUE;
		break;

              case IMGVIEW_DRAG_MODE_ZOOM:
                if((xadj != NULL) && (yadj != NULL) &&
                   (tar_img != NULL) && (src_img != NULL)
		)
                {
		    /* Zoom. */
		    ImgViewZoomItterate(iv, iv->drag_last_y - y);

		    /* Clip bounds, update adjustments and redraw. */
                    ImgViewRealizeChange(iv);
		    ImgViewExposeCB(iv->view_da, NULL, iv);
                }
                status = TRUE;
                break;

	      case IMGVIEW_DRAG_MODE_ZOOM_RECTANGLE:
		if((tar_img != NULL) && (src_img != NULL))
		{
		    GdkEventExpose ev;
		    GdkRectangle *rect;


		    /* Update current rectangle bounds. */
		    iv->drag_zoom_rectangle_cur_x = x;
		    iv->drag_zoom_rectangle_cur_y = y;

		    /* Set up expose event for redraw of only the
		     * rectangular area.
		     */
		    ev.type = GDK_EXPOSE;
		    ev.window = motion->window;
		    ev.count = 0;
		    rect = &ev.area;

		    rect->x = 0;
		    rect->y = 0;

                    rect->width = MAX(iv->drag_zoom_rectangle_cur_x,
                        iv->drag_zoom_rectangle_start_x);
                    rect->height = MAX(iv->drag_zoom_rectangle_cur_y,
                        iv->drag_zoom_rectangle_start_y);

		    /* Redraw. */
/* this dosen't work too well, don't pass an event. */
		    ImgViewExposeCB(iv->view_da, NULL, iv);
		}
		status = TRUE;
		break;
	    }
            /* Reset last drag positions on any motion. */
            iv->drag_last_x = x;
            iv->drag_last_y = y;

	    /* Update last motion event handled time. */
	    iv->last_motion_time = motion->time +
		(gdk_time_get() - motion_timmer);
	    break;
	}

	/* Return event handled status. */
	return(status);
}

/*
 *	View widget "expose_event" callback.
 *
 *	The given event expose can be NULL.
 */
static gint ImgViewExposeCB(
	GtkWidget *widget, GdkEventExpose *expose, gpointer data
)
{
	GdkWindow *window;
	imgview_image_struct *img;
	GtkStyle *style;
	GtkWidget *w = widget;
        imgview_struct *iv = (imgview_struct *)data;
        if((w == NULL) || (iv == NULL))
            return(FALSE);


	if(!iv->initialized || !iv->map_state)
	    return(TRUE);

	if(GTK_WIDGET_NO_WINDOW(w))
	    return(TRUE);
	else
	    window = w->window;
	if(window == NULL)
	    return(TRUE);

	style = gtk_widget_get_style(w);
	if(style == NULL)
	    return(TRUE);

	img = iv->view_img;
	if(img == NULL)
	    return(TRUE);

	/* Blit image. */
	ImgViewBlitView(iv);

        /* Put blitted view image to view drawing area widget's window. */
	if(expose != NULL)
	    ImgViewImageSendRectangle(
                img, (GdkDrawable *)window, style->fg_gc[GTK_STATE_NORMAL],
		&expose->area
            );
	else
	    ImgViewImageSend(
		img, (GdkDrawable *)window, style->fg_gc[GTK_STATE_NORMAL]
	    );

	/* Check drag mode and draw superimposed graphics as needed. */
        /* Drag mode in zoom? */
	if(iv->drag_mode == IMGVIEW_DRAG_MODE_ZOOM)
        {
	    GdkFont	*font = iv->view_font;
            GdkGC       *gc = iv->view_selection_gc;
	    imgview_image_struct *tar_img = iv->view_img;
	    gdouble	zoom = iv->view_zoom;

	    /* Draw zoom value? */
	    if(iv->show_values && (font != NULL) && (gc != NULL) &&
               (tar_img != NULL) && (zoom > 0.0)
	    )
	    {
		gchar s[256];
		gint lbearing, rbearing, width, ascent, descent;

		sprintf(
		    s,
		    "%.0f%%",
		    zoom * 100
		);
		gdk_string_extents(
		    font, s,
		    &lbearing, &rbearing, &width,
		    &ascent, &descent
		);
		gdk_draw_string(
		    (GdkDrawable *)window, font, gc,
		    (tar_img->width / 2) - (width / 2) + lbearing,
		    (tar_img->height / 2) - ((ascent + descent) / 2) + ascent,
		    s
		);
	    }
	}
	/* Drag mode in zoom rectangle? */
	else if(iv->drag_mode == IMGVIEW_DRAG_MODE_ZOOM_RECTANGLE)
	{
	    GdkGC	*gc = iv->view_selection_gc;
	    gint	x0 = iv->drag_zoom_rectangle_start_x,
			x1 = iv->drag_zoom_rectangle_cur_x,
			y0 = iv->drag_zoom_rectangle_start_y,
			y1 = iv->drag_zoom_rectangle_cur_y,
			rw, rh;
            gdouble     zoom = iv->view_zoom;

	    if(x0 > x1)
	    {
		gint t = x1;
		x1 = x0;
		x0 = t;
	    }
            if(y0 > y1)
            {
                gint t = y1;
                y1 = y0;
                y0 = t;
            }

	    rw = x1 - x0;
	    rh = y1 - y0;

	    /* Enough data to draw rectangle? */
	    if((rw > 0) && (rh > 0) && (gc != NULL) && (zoom > 0.0))
	    {
		GdkFont *font = iv->view_font;

		/* Draw `rectangular rubber band'. */
		gdk_draw_rectangle(
		    (GdkDrawable *)window, gc,
		    FALSE,
		    x0, y0, rw, rh
		);
		/* Draw rectangle size? */
		if(iv->show_values && (font != NULL))
		{
		    gchar s[256];
		    gint lbearing, rbearing, width, ascent, descent;

		    sprintf(
			s,
			"%ix%i",
			(gint)((x1 - x0) / zoom),
			(gint)((y1 - y0) / zoom)
		    );
		    gdk_string_extents(
			font, s,
			&lbearing, &rbearing, &width,
			&ascent, &descent
		    );
		    gdk_draw_string(
			(GdkDrawable *)window, font, gc,
			x0 + ((x1 - x0) / 2) - (width / 2) + lbearing,
			y0 + 2 + ascent,
                        s
		    );
		}
	    }
	}

	return(TRUE);
}


/*
 *	Clips the given image viewer's translate bounds, updates the
 *	adjustments, and emits a "changed" signal for the adjustments.
 *
 *	This function should be called whenever changes are made to
 *	the view, ie when the view's translation or zoom has changed or
 *	when a new image has been loaded.
 */
static void ImgViewRealizeChange(imgview_struct *iv)
{
	gdouble zoom;
	gint src_vw, src_vh;
	GtkWidget *w;
	imgview_image_struct *tar_img, *src_img;
	GtkAdjustment *xadj, *yadj;


	if((iv == NULL) ? 1 : !iv->initialized)
	    return;

	tar_img = iv->view_img;
	src_img = iv->orig_img;
	zoom = iv->view_zoom;
	xadj = iv->view_x_adj;
	yadj = iv->view_y_adj;

	/* Must have adjustments. */
	if((xadj == NULL) || (yadj == NULL))
	    return;

	if((tar_img == NULL) || (zoom <= 0.0))
	{
	    xadj->value = 0.0;
	    yadj->value = 0.0;
	    return;
	}

	/* If no source image, then set adjustments to size of
	 * target image.
	 */
	if(src_img == NULL)
	{
	    xadj->value = 0.0;
	    xadj->lower = 0.0;
	    xadj->upper = (gfloat)tar_img->width;
	    xadj->page_increment = 0.0;
	    xadj->page_size = (gfloat)tar_img->width;
            yadj->value = 0.0;
            yadj->lower = 0.0;
            yadj->page_increment = 0.0;
            yadj->upper = (gfloat)tar_img->height;
            yadj->page_size = (gfloat)tar_img->height;
	}
	else
	{
	    /* Calculate the size of the viewed area on the source image. */
	    src_vw = tar_img->width / zoom;
	    src_vh = tar_img->height / zoom;

	    /* Set bounds for adjustments to match source image size. */
	    xadj->lower = 0.0;
	    xadj->upper = src_img->width;
	    yadj->lower = 0.0;
	    yadj->upper = src_img->height;

	    /* Width of viewed source smaller than target width? */
	    if((src_img->width * zoom) < tar_img->width)
	    {
		xadj->value = 0.0;
		xadj->page_size = (gfloat)src_img->width;
	    }
	    else
	    {
		if(xadj->value > (gfloat)(src_img->width - src_vw))
		    xadj->value = (gfloat)(src_img->width - src_vw);
		if(xadj->value < 0.0)
		    xadj->value = 0.0;
		xadj->page_size = (gfloat)src_vw;
		xadj->page_increment = (gfloat)src_vw * xadj->page_size;
	    }

	    /* Height of viewed source smaller than target height? */
	    if((src_img->height * zoom) < tar_img->height)
	    {
		yadj->value = 0.0;
		yadj->page_size = (gfloat)src_img->height;
	    }
	    else
	    {
		if(yadj->value > (gfloat)(src_img->height - src_vh))
		    yadj->value = (gfloat)(src_img->height - src_vh);
		if(yadj->value < 0.0)
		    yadj->value = 0.0;
		yadj->page_size = (gfloat)src_vh;
                yadj->page_increment = (gfloat)src_vh * yadj->page_size;
	    }
        }

	/* Update info label. */
	w = iv->info_label;
	if(w != NULL)
	{
	    gchar text[256];

	    if((src_img == NULL) || !iv->show_values)
		(*text) = '\0';
	    else
		sprintf(
		    text, "%ix%i%s%i%s%i",
		    (gint)(xadj->upper - xadj->lower),
		    (gint)(yadj->upper - yadj->lower),
		    ((gint)xadj->value < 0) ? "" : "+",
		    (gint)xadj->value,
		    ((gint)yadj->value < 0) ? "" : "+",
		    (gint)yadj->value
		);
	    gtk_label_set_text(GTK_LABEL(w), text);
	}

	/* Send "changed" signal to adjustments so the scrollbars are
	 * updated and the view is redrawn.
	 */
	gtk_signal_emit_by_name(GTK_OBJECT(xadj), "changed");
	gtk_signal_emit_by_name(GTK_OBJECT(yadj), "changed");
}

/*
 *	Zoom itteration.
 *
 *	Zooms in or out depending on the value of dz in units
 *	of pixels. Positive dz will zoom out while negative dz will
 *	zoom in.
 *
 *	No bounds will be cliped and view is not redrawn, the calling 
 *	function is responsible for that.
 */
static void ImgViewZoomItterate(
	imgview_struct *iv, gint dz
)
{
	const gdouble	zoom_min = IMGVIEW_ZOOM_MIN,
			zoom_max = IMGVIEW_ZOOM_MAX,
			zoom_rate = IMGVIEW_ZOOM_RATE;
	imgview_image_struct *tar_img;
	GtkAdjustment *xadj, *yadj;
	gint p_vw, p_vh, n_vw, n_vh;
	gdouble	new_zoom,
		prev_zoom = iv->view_zoom;


	if((iv == NULL) ? 1 : !iv->initialized)
	    return;

	if(dz == 0)
	    return;

	xadj = iv->view_x_adj;
	yadj = iv->view_y_adj;
	tar_img = iv->view_img;
	if((xadj == NULL) || (yadj == NULL) || (tar_img == NULL))
	    return;

	/* Update zoom. */
	iv->view_zoom -= ((gfloat)dz * zoom_rate);
	if(iv->view_zoom < zoom_min)
	    iv->view_zoom = zoom_min;
	new_zoom = iv->view_zoom;

	/* Sanitize previous zoom. */
	if(prev_zoom > zoom_max)
	    prev_zoom = zoom_max;
	else if(prev_zoom < zoom_min)
	    prev_zoom = zoom_min;

	/* Calculate previous and new viewable dimensions on the source
	 * image.
	 */
	p_vw = tar_img->width / prev_zoom;
	p_vh = tar_img->height / prev_zoom;
	n_vw = tar_img->width / new_zoom;
	n_vh = tar_img->height / new_zoom;

	/* Adjust translation due to zooming. */
	xadj->value += (p_vw / 2) - (n_vw / 2);
	yadj->value += (p_vh / 2) - (n_vh / 2);
}

/*
 *	Rectangular zoom.
 *
 *	Zooms into the defined rectangle from the given inputs.
 *
 *	No bounds will be cliped and view is not redrawn, the calling
 *      function is responsible for that.
 */
static void ImgViewZoomRectangular(
        imgview_struct *iv,
        gint x0, gint x1,
        gint y0, gint y1
)
{
        gint view_width, view_height;
	gint src_width, src_height;
	gint rw, rh, rcx, rcy;
        imgview_image_struct *tar_img, *src_img;
        GtkAdjustment *xadj, *yadj;
	const gdouble	zoom_min = IMGVIEW_ZOOM_MIN,
			zoom_max = IMGVIEW_ZOOM_MAX;
        gdouble new_zoom,
                prev_zoom = iv->view_zoom;


        if((iv == NULL) ? 1 : !iv->initialized)
            return;

        if((x0 == x1) || (y0 == y1))
            return;

	if(prev_zoom <= 0.0)
	    return;

        xadj = iv->view_x_adj;
        yadj = iv->view_y_adj;
        tar_img = iv->view_img;
	src_img = iv->orig_img;
        if((xadj == NULL) || (yadj == NULL) ||
           (tar_img == NULL) || (src_img == NULL)
	)
            return;

	if((tar_img->width < 1) || (tar_img->height < 1))
	    return;


	/* Flip rectangle coordinates as needed. */
	if(x0 > x1)
	{
	    gint t = x1;
	    x1 = x0;
	    x0 = t;
	}
        if(y0 > y1)
        {
            gint t = y1;
            y1 = y0;
            y0 = t;
        }

	/* Calculate zoomed rectangle size. */
	rw = (x1 - x0) / prev_zoom;
	rh = (y1 - y0) / prev_zoom;

	if((rw < 1) || (rh < 1))
	    return;

        /* Get size of view in window coordinates. */
        view_width = tar_img->width;
        view_height = tar_img->height;

	/* Get size of original image in window coordinates. */
	src_width = src_img->width;
	src_height = src_img->height;

	/* Remove centering offset from rectangle coordinates. */
	x0 += (gint)MIN(((src_width * prev_zoom) - view_width) / 2, 0);
        x1 += (gint)MIN(((src_width * prev_zoom) - view_width) / 2, 0);
        y0 += (gint)MIN(((src_height * prev_zoom) - view_height) / 2, 0);
        y1 += (gint)MIN(((src_height * prev_zoom) - view_height) / 2, 0);

	/* Calculate zoomed rectangle center (be sure to discard offset
	 * used to center image on view if source image is smaller).
	 */
	rcx = (gint)(xadj->value + ((gfloat)x0 / prev_zoom) + ((gfloat)rw / 2));
        rcy = (gint)(yadj->value + ((gfloat)y0 / prev_zoom) + ((gfloat)rh / 2));

	/* Calculate new zoom coefficient. */
	new_zoom = (gdouble)view_width / (gdouble)rw;
	if(((gfloat)view_height / new_zoom) < rh)
	    new_zoom = (gdouble)view_height / (gdouble)rh;

	/* Check if new zoom is out of bounds, if it is then skip. */
	if((new_zoom < zoom_min) || (new_zoom > zoom_max))
	    return;

	/* Update zoom. */
	iv->view_zoom = new_zoom;

	/* Update translation. */
	xadj->value = (gfloat)(rcx - ((gfloat)view_width / new_zoom / 2));
        yadj->value = (gfloat)(rcy - ((gfloat)view_height / new_zoom / 2));
}


/*
 *	Blits the image viewer's orig_img to view_img using the given image
 *	viewer's translation and zoom.
 */
static void ImgViewBlitView(imgview_struct *iv)
{
	GtkAdjustment *xadj, *yadj;
	GtkWidget *w;
	imgview_image_struct *src_img, *tar_img;
	guint depth;			/* Actual depth. */

	gint	src_sk_col,		/* Skips, in * 256. */
		src_sk_row;
	gint	src_cx, src_cy,		/* Cur pos, in * 256. */
		src_x, src_y,		/* Restart pos, in * 256. */
		src_w, src_h,
		src_tw, src_th;
	gint	tar_cx, tar_cy,		/* Cur pos, in * 256. */
		tar_x, tar_y,		/* Restart pos, in * 256. */
		tar_w, tar_h,
		tar_tw, tar_th;
	gdouble zoom;


	if(iv == NULL)
	    return;

	w = iv->view_da;
	if(w == NULL)
	    return;

	xadj = iv->view_x_adj;
	yadj = iv->view_y_adj;
	if((xadj == NULL) || (yadj == NULL))
	    return;

	/* Get pointers to source and target images. */
	src_img = iv->orig_img;
	tar_img = iv->view_img;
	if(tar_img == NULL)
	    return;

        /* Get depth, use target image. */
        depth = ((guint)tar_img->bpp << 3);
        if(depth < 1)
            return;

	/* If no source image then just clear target image. */
	if(src_img == NULL)
	{
	    GtkStyle *style = gtk_widget_get_style(w);
	    if(style != NULL)
	    {
		GdkColor *c = &style->base[GTK_STATE_NORMAL];
		ImgViewImageClear(
		    tar_img,
                    ((guint32)0xff000000) +
                    (((guint32)c->blue & 0x0000ff00) << 8) +
                    (((guint32)c->green & 0x0000ff00)) +
                    (((guint32)c->red & 0x0000ff00) >> 8)
		);
	    }
	    return;
	}


	/* Actual depth of both images must be the same! */
	if(src_img->bpp != tar_img->bpp)
	    return;

	/* Make sure both images have allocated image buffers. */
	if((src_img->mem == NULL) || (tar_img->mem == NULL))
	    return;

	/* Get zoom. */
	zoom = iv->view_zoom;
	if(zoom <= 0.0)
	    return;

	src_sk_col = (gint)(256 / zoom);
	src_sk_row = (gint)(256 / zoom);

	src_tw = (gint)src_img->width * 256;
	src_th = (gint)src_img->height * 256;
	tar_tw = (gint)tar_img->width * 256;
	tar_th = (gint)tar_img->height * 256;

	src_w = (gint)(tar_tw / zoom);
        src_h = (gint)(tar_th / zoom);
	if(src_w > src_tw)
	    src_w = src_tw;
	if(src_h > src_th)
	    src_h = src_th;

	tar_w = (gint)((src_tw * zoom < tar_tw) ? src_tw * zoom : tar_tw);
	tar_h = (gint)((src_th * zoom < tar_th) ? src_th * zoom : tar_th);

	tar_x = (tar_w < tar_tw) ? (tar_tw / 2) - (tar_w / 2) : 0;
	if(tar_x < 0)
	    tar_x = 0;
        tar_y = (tar_h < tar_th) ? (tar_th / 2) - (tar_h / 2) : 0;
	if(tar_y < 0)
	    tar_y = 0;

	tar_w += tar_x;
	tar_h += tar_y;

	src_x = xadj->value * 256;
	src_x = ((src_x + src_w) >= src_tw) ? src_tw - src_w : src_x;
	if(src_x < 0)
	    src_x = 0;
        src_y = yadj->value * 256;
        src_y = ((src_y + src_h) >= src_th) ? src_th - src_h : src_y;
        if(src_y < 0)
	    src_y = 0;

	src_w += src_x;
	src_h += src_y;

	src_cx = src_x;
	src_cy = src_y;
	tar_cx = tar_x;
	tar_cy = tar_y;


        /* Need to clear target image if zoom causes source image to be displayed
	 * smaller than the target image.
	 */
        if(((src_tw * zoom) < tar_tw) || ((src_th * zoom) < tar_th))
        {
            GtkStyle *style = gtk_widget_get_style(w);
            if(style != NULL)
            {
                GdkColor *c = &style->base[GTK_STATE_NORMAL];
                ImgViewImageClear(
                    tar_img,
                    ((guint32)0xff000000) +
                    (((guint32)c->blue & 0x0000ff00) << 8) +
                    (((guint32)c->green & 0x0000ff00)) +
                    (((guint32)c->red & 0x0000ff00) >> 8)
                );
            }
        }

#define DEBUG_CHECK_BOUNDS	\
{ \
 if((src_cx < 0) || ((src_cx >> 8) >= src_img->width)) \
  printf("Source segfault X %i(%i)\n", (src_cx >> 8), src_img->width); \
 if((src_cy < 0) || ((src_cy >> 8) >= src_img->height)) \
  printf("Source segfault Y %i(%i)\n", (src_cy >> 8), src_img->height); \
 if((tar_cx < 0) || ((tar_cx >> 8) >= tar_img->width)) \
  printf("Target segfault X %i(%i)\n", (tar_cx >> 8), tar_img->width); \
 if((tar_cy < 0) || ((tar_cy >> 8) >= tar_img->height)) \
  printf("Target segfault Y %i(%i)\n", (tar_cy >> 8), src_img->height); \
}

	/* 8 bits. */
	if(depth == 8)
	{
	    gint		src_bpp = src_img->bpp,
				tar_bpp = tar_img->bpp;
	    gint		src_bpl = src_img->bpl,
				tar_bpl = tar_img->bpl;
	    const u_int8_t	*src_ptr,
				*src_buf = (const u_int8_t *)src_img->mem;
            u_int8_t		*tar_ptr,
				*tar_buf = (u_int8_t *)tar_img->mem;


            while((src_cy < src_h) &&
                  (tar_cy < tar_h)
            )
            {
/* DEBUG_CHECK_BOUNDS */
                src_ptr = &(src_buf[
                    ((src_cy >> 8) * src_bpl) +
                    ((src_cx >> 8) * src_bpp)
                ]);
                tar_ptr = &(tar_buf[
                    ((tar_cy >> 8) * tar_bpl) +
                    ((tar_cx >> 8) * tar_bpp)
                ]);
                *(u_int8_t *)tar_ptr = *(u_int8_t *)src_ptr;

                /* Increment target x colum. */
                tar_cx += 256;
                /* Go to next target line? */
                if(tar_cx >= tar_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;
        
                    src_cy += src_sk_row;
                    src_cx = src_x;
		    continue;
                }

                /* Increment source x column. */
                src_cx += src_sk_col;
                /* Go to next source line? */
                if(src_cx >= src_w)  
                {
                    tar_cy += 256;
                    tar_cx = tar_x;
         
                    src_cy += src_sk_row;
                    src_cx = src_x;
		    continue;
                }
            }
	}
	/* 15 or 16 bits. */
	else if((depth == 15) || (depth == 16))
	{
	    gint		src_bpp = src_img->bpp,
				tar_bpp = tar_img->bpp;
	    gint		src_bpl = src_img->bpl,
				tar_bpl = tar_img->bpl;
            const u_int8_t      *src_ptr,
                                *src_buf = (const u_int8_t *)src_img->mem;
            u_int8_t            *tar_ptr,
                                *tar_buf = (u_int8_t *)tar_img->mem;


	    while((src_cy < src_h) &&
                  (tar_cy < tar_h)
	    )
	    {
/* DEBUG_CHECK_BOUNDS */
		src_ptr = &(src_buf[
		    ((src_cy >> 8) * src_bpl) +
		    ((src_cx >> 8) * src_bpp)
		]);
		tar_ptr = &(tar_buf[
                    ((tar_cy >> 8) * tar_bpl) +
                    ((tar_cx >> 8) * tar_bpp)
                ]);
		*(u_int16_t *)tar_ptr = *(u_int16_t *)src_ptr;

                /* Increment target x column. */
                tar_cx += 256;
                /* Go to next target line? */
                if(tar_cx >= tar_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;

		    src_cy += src_sk_row;
		    src_cx = src_x;
		    continue;
                }

		/* Increment source x column. */
		src_cx += src_sk_col;
                /* Go to next source line? */
		if(src_cx >= src_w)
		{
                    tar_cy += 256;
                    tar_cx = tar_x;

		    src_cy += src_sk_row;
		    src_cx = src_x;
		    continue;
		}
	    }
	}
	/* 24 bits. */
	else if(depth == 24)
	{
            gint                src_bpp = src_img->bpp,
                                tar_bpp = tar_img->bpp;
            gint                src_bpl = src_img->bpl,
                                tar_bpl = tar_img->bpl;
            const u_int8_t      *src_ptr,
                                *src_buf = (const u_int8_t *)src_img->mem;
            u_int8_t            *tar_ptr,
                                *tar_buf = (u_int8_t *)tar_img->mem;

            while((src_cy < src_h) &&
                  (tar_cy < tar_h)
            )
            {
/* DEBUG_CHECK_BOUNDS */
                src_ptr = &(src_buf[
                    ((src_cy >> 8) * src_bpl) +
                    ((src_cx >> 8) * src_bpp)
                ]);
                tar_ptr = &(tar_buf[
                    ((tar_cy >> 8) * tar_bpl) +
                    ((tar_cx >> 8) * tar_bpp)
                ]);
                *tar_ptr++ = *src_ptr++;
                *tar_ptr++ = *src_ptr++;
                *tar_ptr++ = *src_ptr++;

                /* Increment target x column. */
                tar_cx += 256;
                /* Go to next target line? */
                if(tar_cx >= tar_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;

                    src_cy += src_sk_row;
                    src_cx = src_x;
                    continue;
                }

                /* Increment source x column. */
                src_cx += src_sk_col;
                /* Go to next source line? */
                if(src_cx >= src_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;

                    src_cy += src_sk_row;
                    src_cx = src_x;
                    continue;
                }
            }
        }
	/* 32 bits. */
	else if(depth == 32)
	{
	    gint		src_bpp = src_img->bpp,
				tar_bpp = tar_img->bpp;
            gint                src_bpl = src_img->bpl,
                                tar_bpl = tar_img->bpl;
            const u_int8_t      *src_ptr,
                                *src_buf = (const u_int8_t *)src_img->mem;
            u_int8_t            *tar_ptr,
                                *tar_buf = (u_int8_t *)tar_img->mem;

/* printf("Blit 32\n"); */
            while((src_cy < src_h) &&
                  (tar_cy < tar_h)
            )
            {
/* DEBUG_CHECK_BOUNDS */
                src_ptr = &(src_buf[
                    ((src_cy >> 8) * src_bpl) +
                    ((src_cx >> 8) * src_bpp)
                ]);
                tar_ptr = &(tar_buf[
                    ((tar_cy >> 8) * tar_bpl) +
                    ((tar_cx >> 8) * tar_bpp)
                ]);
                *(u_int32_t *)tar_ptr = *(u_int32_t *)src_ptr;


                /* Increment target x column. */
                tar_cx += 256;
                /* Go to next target line? */
                if(tar_cx >= tar_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;

                    src_cy += src_sk_row;
                    src_cx = src_x;
		    continue;
                }

                /* Increment source x column. */
                src_cx += src_sk_col;
                /* Go to next source line? */
                if(src_cx >= src_w)
                {
                    tar_cy += 256;
                    tar_cx = tar_x;

                    src_cy += src_sk_row;
                    src_cx = src_x;
		    continue;
                }
            }
	}

#undef DEBUG_CHECK_BOUNDS
}

/*
 *	Updates the contents of the WM icon (if any) on the given
 *	image viewer. If the image viewer's toplevel_is_window is FALSE
 *	then there will be no WM icon to update.
 *
 *	If iv->show_image_on_wm_icon is FALSE then this call has no
 *	affect.
 */
static void ImgViewWMIconUpdate(imgview_struct *iv)
{
	const gint	width = IMGVIEW_WM_ICON_WIDTH,
			height = IMGVIEW_WM_ICON_HEIGHT;
	gint x, y;
	GdkWindow *window;
	GdkBitmap *mask = NULL;
	GdkPixmap *pixmap = NULL;
        GtkWidget *w;
	GtkStyle *style;
	gchar *bm_data, *bm_ptr;
        imgview_image_struct *src_img;


        if((iv == NULL) ? 1 : !iv->initialized)
            return;

	/* Skip if image viewer's toplevel widget is not a GtkWindow or
	 * we do not want to show image on wm icon.
	 */
	if(!iv->toplevel_is_window || !iv->show_image_on_wm_icon)
	    return;

	w = iv->toplevel;
	if((w == NULL) ? 1 : GTK_WIDGET_NO_WINDOW(w))
	    return;
	else
	    window = w->window;

	if((width < 1) || (height < 1))
	    return;


	/* Calculate size of bitmap in bits. */
/*
	if(width % 8)
	    bm_width = (gint)(width / 8) + 1;
	else
	    bm_width = (gint)(width / 8) + 1;
	bm_height = height;
 */

	/* Allocate bitmap data. */
	bm_data = (gchar *)g_malloc(width * height);
	if(bm_data != NULL)
	{
	    /* Set bitmap data. */
	    for(y = 0; y < height; y++)
	    {
		for(x = 0; x < width; x++)
		{
		    bm_ptr = &(bm_data[
			(y * width) + x
		    ]);
		    *bm_ptr = 0xff;
		}
	    }

	    /* Create new WM icon mask. */
	    mask = gdk_bitmap_create_from_data(
		window, bm_data, width, height
	    );

	    /* Deallocate bitmap data. */
	    g_free(bm_data);
	    bm_data = NULL;
	}




	/* Create new WM icon pixmap. */
	pixmap = gdk_pixmap_new(
	    window, width, height, -1
	);
	src_img = iv->orig_img;
	style = gtk_widget_get_style(w);
	if(style != NULL)
	{
	    /* Create tempory target image. */
	    imgview_image_struct *tar_img = ImgViewImageNew(
		width, height, 0
	    );

	    /* Clear pixmap if source image is NULL. */
	    if((src_img == NULL) && (tar_img != NULL))
	    {
		/* No source image, so just clear it. */
		ImgViewImageClear(tar_img, 0xffffffff);
	    }
	    else if(tar_img != NULL)
	    {
		/* Got source image, now copy and resize the source image
		 * buffer to the target image.
		 */
		GUIResizeBuffer(
		    tar_img->bpp,		/* Tar/src bpp should be same. */
		    src_img->mem,		/* Source buffer. */
		    src_img->width, src_img->height, src_img->bpl,
		    tar_img->mem,		/* Target buffer. */
		    tar_img->width, tar_img->height, tar_img->bpl
		);
	    }

	    /* Put target image to the WM icon pixmap. */
	    ImgViewImageSend(
		tar_img,
		(GdkDrawable *)pixmap, 
		style->fg_gc[GTK_STATE_NORMAL]
	    );

	    /* Delete tempory target image. */
	    gdk_flush();
	    ImgViewImageDelete(tar_img);
	}


        /* Update WM icon for our toplevel window's GdkWindow. */
	if((mask != NULL) && (pixmap != NULL))
	    gdk_window_set_icon(
		window,
		NULL,		/* WM icon GdkWindow. */
		pixmap,		/* WM icon GdkPixmap. */
		mask		/* WM icon GdkBitmap. */
	    );
	/* Flush output to make sure icon is changed. */
	gdk_flush();

        /* Unref old icon mask and pixmap. */
        if(iv->wm_icon_mask != NULL)
            gdk_bitmap_unref(iv->wm_icon_mask);
        if(iv->wm_icon_pixmap != NULL)
            gdk_pixmap_unref(iv->wm_icon_pixmap);

	/* Record new mask and pixmap. */
	iv->wm_icon_mask = mask;
	iv->wm_icon_pixmap = pixmap;

}


/*
 *	Recreates the image view buffer by checking the visual and the size of
 *	the view translation and geometry.
 *
 *	This function should be called on resizes and zoom changes on the view
 *	drawing area widget.
 *
 *	On failure, the iv's view_img will be NULL.
 */
static void ImgViewBufferRecreate(imgview_struct *iv)
{
	gint width, height;
	GtkWidget *w;
	imgview_image_struct *img;


        if((iv == NULL) ? 1 : !iv->initialized)
            return;


	/* Destroy old image. */
	if(iv->view_img != NULL)
	{
	    gdk_flush();
	    ImgViewImageDelete(iv->view_img);
	    iv->view_img = NULL;
	}

	/* Get view drawing area and visual (both must be valid). */
	w = iv->view_da;
	if(w == NULL)
	    return;

	/* Get size that we need to recreate the view buffer image as. */
	width = w->allocation.width;
	height = w->allocation.height;
	if((width < 1) || (height < 1))
	    return;

	/* Create new view buffer image. */
	iv->view_img = img = ImgViewImageNew(
	    width, height, 0
	);
	if(img == NULL)
	    return;
}


/*
 *	Returns TRUE if the image viewer's toplevel widget is a
 *	GtkWindow otherwise it is a GtkVBox.
 */
gbool ImgViewToplevelIsWindow(imgview_struct *iv)
{
        if((iv == NULL) ? 1 : !iv->initialized)
            return(FALSE);

	return(iv->toplevel_is_window);
}

/*
 *	Returns the toplevel widget, this may be either a GtkVBox or
 *	GtkWindow depending on what iv->toplevel_is_window says.
 */
GtkWidget *ImgViewGetToplevelWidget(imgview_struct *iv)
{
	if((iv == NULL) ? 1 : !iv->initialized)
            return(NULL);

        return(iv->toplevel);
}

/*
 *	Returns the view drawing area widget.
 */
GtkDrawingArea *ImgViewGetViewWidget(imgview_struct *iv)
{
	if((iv == NULL) ? 1 : !iv->initialized)
	    return(NULL);

	return((GtkDrawingArea *)iv->view_da);
}

/*
 *	Returns the menu widget.
 */
GtkMenu *ImgViewGetMenuWidget(imgview_struct *iv)
{
        if((iv == NULL) ? 1 : !iv->initialized)
            return(NULL);

        return((GtkMenu *)iv->menu);
}

/*
 *	Deallocates the orig_img on the image viewer and redraws.
 *
 *	This basically unloads the image.
 */
void ImgViewClear(imgview_struct *iv)
{
	GtkAdjustment *adj;


	if((iv == NULL) ? 1 : !iv->initialized)
            return;

	if(iv->orig_img != NULL)
	{
	    gdk_flush();
	    ImgViewImageDelete(iv->orig_img);
	    iv->orig_img = NULL;
	}

        /* Reset zoom. */
        iv->view_zoom = 1.0;

	/* Reset translations. */
        adj = iv->view_x_adj;
	if(adj != NULL)
	{
            adj->value = 0.0;
	}
        adj = iv->view_y_adj;
	{
            adj->value = 0.0;
	}

	/* Update WM icon as needed. */
	ImgViewWMIconUpdate(iv);

        /* Clip bounds, update adjustments, and redraw. */
        ImgViewRealizeChange(iv);
	ImgViewUpdateMenus(iv);
	ImgViewExposeCB(iv->view_da, NULL, iv);
}

/*
 *	Loads the given image data to the image viewer. The given image data
 *	buffer format must match the format specified by format.
 *
 *	If an image is already loaded then it will be unloaded.
 *
 *	Returns non-zero on error.
 */
gint ImgViewLoad(
	imgview_struct *iv,
	gint width, gint height,
	gint bytes_per_line,	/* Can be 0 to auto calculate. */
	gint format,		/* One of IMGVIEW_FORMAT_*. */
	const guint8 *data
)
{
	GdkColormap *colormap;
	imgview_image_struct *img;
	GtkWidget *w;


	if(iv == NULL)
	    return(-1);

	/* Unload image first and reset translations as needed. */
	ImgViewClear(iv);

	/* If given data is NULL, then give up. */
	if((data == NULL) || (width < 1) || (height < 1))
	    return(-1);

	/* Get colormap. */
	w = iv->view_da;
	if(w != NULL)
	    colormap = gtk_widget_get_colormap(w);
	else
	    colormap = NULL;
	if(colormap == NULL)
	    return(-3);

	/* Create new image. */
        iv->orig_img = img = ImgViewImageNew(
	    width, height, 0
	);
	if(img == NULL)
	    return(-1);

	/* Copy given data to target image if target image's buffer
	 * is allocated. Both source image data and target image data
	 * must be of the same width and height but may differ in
	 * bytes per line and/or bytes per pixel.
	 */
	if(img->mem != NULL)
        {
            gint x, y, bc, bc_min;
            gint src_bpp, src_bpl;
	    gint	tar_bpp = img->bpp,
			tar_bpl = img->bpl;
	    const u_int8_t	*src_ptr;
	    u_int8_t		*tar_ptr,
				*tar_buf = (u_int8_t *)img->mem;

	    /* Check source image data format, to calculate source image
	     * data bytes per line and bytes per pixel units.
	     */
	    switch(format)
	    {
	      case IMGVIEW_FORMAT_RGBA:
		src_bpp = 4;
		break;
              case IMGVIEW_FORMAT_RGB:
                src_bpp = 3;
                break;
              case IMGVIEW_FORMAT_GREYSCALE32:
                src_bpp = 4;
                break;
              case IMGVIEW_FORMAT_GREYSCALE16:
                src_bpp = 2;
                break;
              case IMGVIEW_FORMAT_GREYSCALE8:
                src_bpp = 1;
                break;
	      default:
		/* Unsupported source image format, safest is to assume
		 * lowest bytes per pixel unit.
		 */
		src_bpp = 1;
		break;
	    }
	    /* Calculate source image bytes per line unit, if given
	     * bytes_per_line is 0 then that means they want us to 
	     * calculate it.
	     */
	    src_bpl = (gint)((bytes_per_line == 0) ?
		width * src_bpp : bytes_per_line
	    );

	    /* Calculate byte count minimum, the smaller of the two
	     * bytes per pixel units.
	     */
	    bc_min = MIN(src_bpp, tar_bpp);

	    /* Itterate through each row. */
	    for(y = 0; y < height; y++)
	    {
		for(x = 0; x < width; x++)
		{
		    /* Get source and target 8 bit pointers at start of
		     * of current pixel.
		     */
		    src_ptr = &(data[
			(y * src_bpl) + (x * src_bpp)
		    ]);
		    tar_ptr = &(tar_buf[
			(y * tar_bpl) + (x * tar_bpp)
		    ]);

		    /* Copy bytes from current source pixel to current
		     * target pixel, incrementing each pointer one byte
		     * after each byte is copied.
		     */
		    for(bc = 0; bc < bc_min; bc++)
			*tar_ptr++ = *src_ptr++;

		    /* Clear any remaining bytes on the current
		     * target pixel.
		     */
/* Note this will cause problems with greyscale source data. */
		    for(; bc < tar_bpp; bc++)
			*tar_ptr++ = 0xff;
                }
            }	/* Itterate through each row. */

        }

        /* Update WM icon as needed. */
        ImgViewWMIconUpdate(iv);

	/* Clip bounds, update adjustments, and redraw. */
	ImgViewRealizeChange(iv);
	ImgViewUpdateMenus(iv);
	ImgViewExposeCB(iv->view_da, NULL, iv);

	return(0);
}

/*
 *	Creates a new image viewer.
 */
imgview_struct *ImgViewNew(
	gbool show_toolbar,
	gbool show_values,
	gbool show_image_on_wm_icon,
	gbool toplevel_is_window,
	GtkWidget **toplevel_rtn
)
{
	gint	border_minor = 2;
	GdkColormap *colormap;
	GtkAdjustment *adj;
	GdkFont *font;
	GdkColor *c;
	GtkRcStyle *rcstyle;
	GtkStyle *style;
	GtkWidget *w, *parent, *parent2, *parent3, *parent4, *parent5;
	GtkAccelGroup *accelgrp;
	imgview_struct *iv = (imgview_struct *)calloc(
	    1, sizeof(imgview_struct)
	);


        if(toplevel_rtn != NULL)
            (*toplevel_rtn) = NULL;

gdk_rgb_init();

	if(iv == NULL)
	     return(NULL);

	/* Reset values. */
	iv->initialized = TRUE;
	iv->map_state = FALSE;
	iv->toplevel_is_window = toplevel_is_window;
	iv->show_values = show_values;
	iv->show_image_on_wm_icon = show_image_on_wm_icon;
	iv->view_zoom_toid = (guint)-1;


	/* Cursors. */
	iv->translate_cur = gdk_cursor_new(GDK_FLEUR);
	iv->zoom_cur = gdk_cursor_new(GDK_SIZING);
	iv->zoom_rectangle_cur = gdk_cursor_new(GDK_CROSS);


        /* Keyboard accelerator group. */
	iv->accelgrp = accelgrp = gtk_accel_group_new();


	/* View adjustments. */
	iv->view_x_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0,                /* Current value. */
            0.0, 0.0,           /* Lower, upper. */
            10.0, 50.0,         /* Step inc, page inc. */
            0.0                 /* Page size. */
        );
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(ImgViewAdjustmentChangedCB),
            (gpointer)iv
        );

        iv->view_y_adj = adj = (GtkAdjustment *)gtk_adjustment_new(
            0.0,		/* Current value. */
            0.0, 0.0,		/* Lower, upper. */
            10.0, 50.0,		/* Step inc, page inc. */
	    0.0			/* Page size. */
        );
        gtk_signal_connect(
            GTK_OBJECT(adj), "value_changed",
            GTK_SIGNAL_FUNC(ImgViewAdjustmentChangedCB),
            (gpointer)iv
        );


        /* Check if toplevel is to be created as a GtkWindow. */
	if(toplevel_is_window)
	{
	    GdkWindow *window = NULL;

	    /* Create a toplevel GtkWindow with RGB buffers. */
            gtk_widget_push_visual(gdk_rgb_get_visual());
            gtk_widget_push_colormap(gdk_rgb_get_cmap());
	    iv->toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
            gtk_widget_pop_visual();
            gtk_widget_pop_colormap();
            gtk_window_set_policy(
                GTK_WINDOW(w), TRUE, TRUE, FALSE
            );
/* Do not set usize, so calling function can set view_da's usize
 * for better image fitting.
 */
/*	    gtk_widget_set_usize(w, 320, 240); */
	    gtk_widget_realize(w);
	    gtk_window_set_title(GTK_WINDOW(w), IMGVIEW_TITLE);
	    if(!GTK_WIDGET_NO_WINDOW(w))
	    {
		GdkGeometry geo;


		window = w->window;	/* Update toplevel GdkWindow pointer. */

		geo.min_width = IMGVIEW_MIN_WIDTH;
		geo.min_height = IMGVIEW_MIN_HEIGHT;
		geo.base_width = 0;
		geo.base_height = 0;
		geo.width_inc = 1;
		geo.height_inc = 1;
/*
		geo.min_aspect = 1.3;
		geo.max_aspect = 1.3;
 */
		gdk_window_set_geometry_hints(
		    window,
		    &geo,
		    GDK_HINT_MIN_SIZE | GDK_HINT_BASE_SIZE |
		    /* GDK_HINT_ASPECT | */
		    GDK_HINT_RESIZE_INC
		);

		if(!show_image_on_wm_icon)
		    GUISetWMIcon(window, (u_int8_t **)iv_xpm);
	    }
            gtk_signal_connect(
                GTK_OBJECT(w), "delete_event",
                GTK_SIGNAL_FUNC(ImgViewCloseCB),
                (gpointer)iv
            );
            gtk_signal_connect(
                GTK_OBJECT(w), "destroy",
                GTK_SIGNAL_FUNC(ImgViewDestroyCB),
                (gpointer)iv
            );
	    gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	    /* Update toplevel return. */
            if(toplevel_rtn != NULL)
                (*toplevel_rtn) = w;
	    parent = w;

	    /* Main vbox in window. */
            iv->main_vbox = w = gtk_vbox_new(FALSE, border_minor);
	    gtk_container_add(GTK_CONTAINER(parent), w);
	    gtk_widget_show(w);
	    parent = w;
	}
	else
	{
	    /* Create a vbox as the toplevel, which will be parented by
	     * the calling function.
	     */
	    iv->toplevel = w = gtk_vbox_new(FALSE, border_minor);
            gtk_signal_connect(
                GTK_OBJECT(w), "destroy",
                GTK_SIGNAL_FUNC(ImgViewDestroyCB),
                (gpointer)iv
            );
            gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	    parent = w;

	    iv->main_vbox = NULL;

            /* Update toplevel return. */
            if(toplevel_rtn != NULL)
                (*toplevel_rtn) = w;
	}

	/* New parent is set to either a GtkWindow or a GtkVBox depending
	 * on toplevel_is_window.
	 */

	/* Toolbar. */
	iv->toolbar_map_state = show_toolbar;
	/* Toolbar toplevel hbox. */
	iv->toolbar_toplevel = w = gtk_hbox_new(FALSE, border_minor);
	gtk_widget_set_usize(w, -1, IMGVIEW_TOOLBAR_HEIGHT);
        gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	if(show_toolbar)
	    gtk_widget_show(w);
	parent3 = w;

	/* Toolbar depressed frame. */
	w = gtk_frame_new(NULL);
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
        gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
        parent4 = w;
	/* Hbox inside depressed frame. */
	w = gtk_hbox_new(FALSE, border_minor);
	gtk_container_add(GTK_CONTAINER(parent4), w);
	gtk_widget_show(w);
	parent4 = w;

	/* Toolbar info label. */
	w = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(parent4), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
        parent5 = w;
	iv->info_label = w = gtk_label_new("");
	gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(parent5), w, FALSE, FALSE, border_minor);
	gtk_widget_show(w);

	/* Toolbar buttons. */
	iv->zoom_in_btn = w = GUIButtonPixmap(
	    (u_int8_t **)icon_zoom_in_16x16_xpm
	);
        gtk_widget_set_usize(w, 16, 16);
        gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
        gtk_widget_set_name(w, IMGVIEW_WNAME_BUTTON_ZOOM_IN);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "pressed",
            GTK_SIGNAL_FUNC(ImgViewZoomInPressedCB),
            (gpointer)iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "released",
            GTK_SIGNAL_FUNC(ImgViewZoomReleasedCB),
            (gpointer)iv
        );
        gtk_widget_show(w);

        iv->zoom_out_btn = w = GUIButtonPixmap(
            (u_int8_t **)icon_zoom_out_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
        gtk_widget_set_name(w, IMGVIEW_WNAME_BUTTON_ZOOM_OUT);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "pressed",
            GTK_SIGNAL_FUNC(ImgViewZoomOutPressedCB),
            (gpointer)iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "released",
            GTK_SIGNAL_FUNC(ImgViewZoomReleasedCB),
            (gpointer)iv
        );
        gtk_widget_show(w);

        iv->zoom_onetoone_btn = w = GUIButtonPixmap(
            (u_int8_t **)icon_zoom_onetoone_16x16_xpm
        );
	gtk_widget_set_usize(w, 16, 16);
	gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
        gtk_widget_set_name(w, IMGVIEW_WNAME_BUTTON_ZOOM_ONETOONE);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(ImgViewZoomOneToOneCB),
            (gpointer)iv
        );
        gtk_widget_show(w);

        iv->zoom_tofit_btn = w = GUIButtonPixmap(
            (u_int8_t **)icon_zoom_tofit_16x16_xpm
        );
        gtk_widget_set_usize(w, 16, 16);
        gtk_button_set_relief(GTK_BUTTON(w), GTK_RELIEF_NONE);
        gtk_widget_set_name(w, IMGVIEW_WNAME_BUTTON_ZOOM_TOFIT);
        gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
        gtk_signal_connect(
            GTK_OBJECT(w), "clicked",
            GTK_SIGNAL_FUNC(ImgViewZoomToFitCB),
            (gpointer)iv
        );
        gtk_widget_show(w);

	/* Show toolbar? */
	if(show_toolbar)
	    gtk_widget_show(iv->toolbar_toplevel);

        /* Table to hold drawing area and scrollbars. */
        w = gtk_table_new(2, 2, FALSE);
        gtk_table_set_row_spacing(GTK_TABLE(w), 0, border_minor);
        gtk_table_set_col_spacing(GTK_TABLE(w), 0, border_minor);
        gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
        gtk_widget_show(w);
        parent2 = w;

	/* Frame for drawing area. */
        w = gtk_frame_new(NULL);
        gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_IN);
        gtk_table_attach(
            GTK_TABLE(parent2), w,
            0, 1, 0, 1,
            GTK_EXPAND | GTK_SHRINK | GTK_FILL,
            GTK_EXPAND | GTK_SHRINK | GTK_FILL,
            0, 0
        );
        gtk_widget_show(w);
	parent3 = w;

	/* View drawing area (with RGB buffers if toplevel is not a
	 * GtkWindow).
	 */
	if(!toplevel_is_window)
	{
	    gtk_widget_push_visual(gdk_rgb_get_visual());
	    gtk_widget_push_colormap(gdk_rgb_get_cmap());
	}
	iv->view_da = w = gtk_drawing_area_new();
	if(!toplevel_is_window)
	{
	    gtk_widget_pop_visual();
	    gtk_widget_pop_colormap();
	}
        gtk_widget_set_name(w, IMGVIEW_WNAME_VIEW);
	/* Must be able to get keyboard focus. */
	GTK_WIDGET_SET_FLAGS(
	    w,
	    GTK_CAN_FOCUS | GTK_CAN_DEFAULT
	);
	/* Set view style. */
	rcstyle = gtk_rc_style_new();
        rcstyle->color_flags[GTK_STATE_NORMAL] = GTK_RC_BASE;
	rcstyle->color_flags[GTK_STATE_SELECTED] = GTK_RC_BASE;
        rcstyle->color_flags[GTK_STATE_INSENSITIVE] = GTK_RC_BASE;
	c = &rcstyle->base[GTK_STATE_NORMAL];
	c->red = 1.0 * (guint16)-1;
        c->green = 1.0 * (guint16)-1;
        c->blue = 1.0 * (guint16)-1;
        c = &rcstyle->base[GTK_STATE_SELECTED];
        c->red = 0.0 * (guint16)-1;
        c->green = 0.0 * (guint16)-1;
        c->blue = 0.61 * (guint16)-1;
        c = &rcstyle->base[GTK_STATE_INSENSITIVE];
        c->red = 1.0 * (guint16)-1;
        c->green = 1.0 * (guint16)-1;
        c->blue = 1.0 * (guint16)-1;
	gtk_widget_modify_style(w, rcstyle);
	GUIRCStyleDeallocUnref(rcstyle);
	/* Needs to accept focus for keyboard modifier events. */
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_FOCUS);
        gtk_widget_add_events(
            w,
            GDK_EXPOSURE_MASK | GDK_ENTER_NOTIFY_MASK |
	    GDK_LEAVE_NOTIFY_MASK | GDK_KEY_PRESS_MASK |
            GDK_KEY_RELEASE_MASK | GDK_BUTTON_PRESS_MASK |
            GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
            GDK_POINTER_MOTION_HINT_MASK | GDK_STRUCTURE_MASK
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "enter_notify_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "leave_notify_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "key_press_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
	);
        gtk_signal_connect(
            GTK_OBJECT(w), "key_release_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_press_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "button_release_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "motion_notify_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "expose_event",
            GTK_SIGNAL_FUNC(ImgViewExposeCB),
            iv
        );
        gtk_signal_connect(
            GTK_OBJECT(w), "configure_event",
            GTK_SIGNAL_FUNC(ImgViewViewEventCB),
            iv
        );
	gtk_container_add(GTK_CONTAINER(parent3), w);
	gtk_widget_show(w);

	/* Reset view image buffer to NULL, it will be created when
	 * ImgViewBufferRecreate() is called for the first time.
	 */
	iv->view_img = NULL;

	/* Get style from info label and use its font. */
	style = gtk_widget_get_style(iv->info_label);
	font = ((style != NULL) ? style->font : NULL);
        /* Create view selection cursory GC. */
	colormap = gtk_widget_get_colormap(w);
	if(colormap == NULL)
	    colormap = gdk_colormap_get_system();
	if(colormap != NULL)
	{
	    GdkColor *cf, *cb;
            GdkGCValues gcv;
	    GdkWindow *window = (GdkWindow *)GDK_ROOT_PARENT();
            GdkGCValuesMask gcv_mask = (
		GDK_GC_FOREGROUND | GDK_GC_BACKGROUND | GDK_GC_FONT |
		GDK_GC_FUNCTION | GDK_GC_FILL | GDK_GC_LINE_WIDTH |
		GDK_GC_LINE_STYLE | GDK_GC_CAP_STYLE | GDK_GC_JOIN_STYLE
	    );
            cf = &gcv.foreground;
            cf->red = (guint16)(0.0 * (guint16)-1);
            cf->green = (guint16)(0.0 * (guint16)-1);
            cf->blue = (guint16)(0.0 * (guint16)-1);
            gdk_colormap_alloc_color(colormap, cf, TRUE, TRUE);
	    cb = &gcv.background;
            cb->red = (guint16)(1.0 * (guint16)-1);
            cb->green = (guint16)(1.0 * (guint16)-1);
            cb->blue = (guint16)(1.0 * (guint16)-1);
            gdk_colormap_alloc_color(colormap, cb, TRUE, TRUE);
	    if(font != NULL)
		gdk_font_ref(font);
	    gcv.font = font;
            gcv.function = GDK_INVERT;
            gcv.fill = GDK_SOLID;
            gcv.line_width = 1;
            gcv.line_style = GDK_LINE_SOLID;
            gcv.cap_style = GDK_CAP_NOT_LAST;
            gcv.join_style = GDK_JOIN_MITER;
            iv->view_selection_gc = gdk_gc_new_with_values(
		window, &gcv, gcv_mask
	    );
            gdk_colormap_free_colors(colormap, cf, 1);
	    gdk_colormap_free_colors(colormap, cb, 1);
            if(font != NULL)
                gdk_font_unref(font);
	}
	/* Record view font if available, be sure to add a ref count
	 * to it since we want to keep it around.
	 */
	if(font != NULL)
	{
	    gdk_font_ref(font);
	    iv->view_font = font;
	}


	/* Create scroll bars, make sure that we increase the
	 * ref count for the adjustment since it will be unref'ed
	 * when the scroll bar is destroyed.
	 */
	gtk_object_ref(GTK_OBJECT(iv->view_x_adj));
        w = gtk_hscrollbar_new(iv->view_x_adj);
        gtk_table_attach(
            GTK_TABLE(parent2), w,
            0, 1, 1, 2,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL,
            GTK_FILL,
            0, 0
        );
        gtk_widget_show(w);

	gtk_object_ref(GTK_OBJECT(iv->view_y_adj));
        w = gtk_vscrollbar_new(iv->view_y_adj);
        gtk_table_attach(
            GTK_TABLE(parent2), w,
            1, 2, 0, 1,
            GTK_FILL,
            GTK_EXPAND | GTK_SHRINK | GTK_FILL,
            0, 0
        );
        gtk_widget_show(w);



	/* Right-click menu. */
	iv->menu = w = (GtkWidget *)GUIMenuCreate();
	if(w != NULL)
	{
	    GtkWidget *menu = w;
	    GtkWidget *fw;
	    gint accel_key;
	    gpointer accel_group = NULL;
	    guint accel_mods;
	    u_int8_t **icon;
	    const gchar *label = NULL;
	    gpointer mclient_data = iv;
	    void (*func_cb)(GtkWidget *w, gpointer) = NULL;

#define DO_ADD_MENU_ITEM_LABEL	\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accel_group, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_ITEM_CHECK	\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_CHECK, accel_group, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_SEP		\
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}
/*
            icon = (u_int8_t **)icon_zoom_in_20x20_xpm;
            label = "Zoom In";
            accel_key = 0;
            accel_mods = 0;
	    func_cb = ImgViewZoomInCB;
            DO_ADD_MENU_ITEM_LABEL
            iv->zoom_in_mi = w;

	    icon = (u_int8_t **)icon_zoom_out_20x20_xpm;
	    label = "Zoom Out";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = ImgViewZoomOutCB;
	    DO_ADD_MENU_ITEM_LABEL
	    iv->zoom_out_mi = w;
 */
            icon = (u_int8_t **)icon_zoom_onetoone_20x20_xpm;
            label = "Zoom 1:1";
            accel_key = 0;
            accel_mods = 0;
	    func_cb = ImgViewZoomOneToOneCB;
            DO_ADD_MENU_ITEM_LABEL
            iv->zoom_onetoone_mi = w;

            icon = (u_int8_t **)icon_zoom_tofit_20x20_xpm;
            label = "Zoom To Fit";
            accel_key = 0;
            accel_mods = 0;
	    func_cb = ImgViewZoomToFitCB;
            DO_ADD_MENU_ITEM_LABEL
            iv->zoom_tofit_mi = w;

	    DO_ADD_MENU_SEP

            icon = NULL;
            label = "ToolBar";
            accel_key = 0;
            accel_mods = 0;
            func_cb = ImgViewToolBarToggleCB;
            DO_ADD_MENU_ITEM_CHECK
            iv->show_toolbar_micheck = w;
	    GTK_CHECK_MENU_ITEM(w)->active = iv->toolbar_map_state;

            icon = NULL;
            label = "Values";
            accel_key = 0;
            accel_mods = 0;
            func_cb = ImgViewValuesToggleCB;
            DO_ADD_MENU_ITEM_CHECK
            iv->show_values_micheck = w;
            GTK_CHECK_MENU_ITEM(w)->active = iv->show_values;

#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_ITEM_CHECK
#undef DO_ADD_MENU_SEP

	}

	/* Reset and update menus. */
	ImgViewReset(iv, FALSE);

	return(iv);
}

/*
 *	Deallocates all resources on the given image viewer.
 *
 *	If repeating view zoom is active then it will be removed.
 *	Any loaded image indicated by orig_img will be unloaded and
 *	orig_img reset.
 */
void ImgViewReset(imgview_struct *iv, gbool need_unmap)
{
        if((iv == NULL) ? 1 : !iv->initialized)
            return;

	/* Reset zoom timeout as needed. */
	if(iv->view_zoom_toid != (guint)-1)
	{
	    gtk_timeout_remove(iv->view_zoom_toid);
	    iv->view_zoom_toid = (guint)-1;
	}

	/* Clear drag mode. */
	iv->drag_mode = IMGVIEW_DRAG_MODE_NONE;

	/* Reset last drag position. */
	iv->drag_last_x = 0;
	iv->drag_last_y = 0;

	/* Reset last drag rectangular position. */
        iv->drag_zoom_rectangle_start_x =
            iv->drag_zoom_rectangle_cur_x = 0;
        iv->drag_zoom_rectangle_start_y =
            iv->drag_zoom_rectangle_cur_y = 0;

	/* Reset last motion event time. */
	iv->last_motion_time = 0;

	/* Unload current image (if any) and update menus. */
	ImgViewClear(iv);

	if(need_unmap)
	    ImgViewUnmap(iv);
}

/*
 *	Updates menus on the image viewer to reflect its current states.
 */
void ImgViewUpdateMenus(imgview_struct *iv)
{
	static gbool reenterant = FALSE;
	gbool sensitivity, state;
	GtkWidget *w;
	imgview_image_struct *src_img;
        if((iv == NULL) ? 1 : !iv->initialized)
            return;

	if(reenterant)
	    return;
	else
	    reenterant = TRUE;

	src_img = iv->orig_img;

#define DO_SET_SENSITIVITY	\
{ \
 if(w != NULL) \
  gtk_widget_set_sensitive(w, sensitivity); \
}

#define DO_SET_CHECK_MENU_ITEM	\
{ \
 if(w != NULL) \
  GTK_CHECK_MENU_ITEM(w)->active = state; \
}

	/* Update toolbar widgets. */
	w = iv->zoom_in_btn;
	sensitivity = ((src_img != NULL) ? TRUE : FALSE);
	DO_SET_SENSITIVITY
        w = iv->zoom_out_btn;
        sensitivity = ((src_img != NULL) ? TRUE : FALSE);
        DO_SET_SENSITIVITY
        w = iv->zoom_onetoone_btn;
        sensitivity = ((src_img != NULL) ? TRUE : FALSE);
        DO_SET_SENSITIVITY
        w = iv->zoom_tofit_btn;
        sensitivity = ((src_img != NULL) ? TRUE : FALSE);
        DO_SET_SENSITIVITY

        /* Update menu items. */
        w = iv->zoom_in_mi;
        sensitivity = ((src_img != NULL) ? TRUE : FALSE);
        DO_SET_SENSITIVITY
        w = iv->zoom_out_mi;
        sensitivity = ((src_img != NULL) ? TRUE : FALSE);
        DO_SET_SENSITIVITY
        w = iv->zoom_onetoone_mi;
        sensitivity = ((src_img != NULL) ? TRUE : FALSE);
        DO_SET_SENSITIVITY
        w = iv->zoom_tofit_mi;
        sensitivity = ((src_img != NULL) ? TRUE : FALSE);
        DO_SET_SENSITIVITY

	w = iv->show_toolbar_micheck;	
	state = iv->toolbar_map_state;
	DO_SET_CHECK_MENU_ITEM
        w = iv->show_values_micheck;
        state = iv->show_values;
        DO_SET_CHECK_MENU_ITEM


#undef DO_SET_SENSITIVITY
#undef DO_SET_CHECK_MENU_ITEM

	reenterant = FALSE;
}


/*
 *	Shows or hides the tool bar.
 */
void ImgViewShowToolBar(imgview_struct *iv, gbool show)
{
        if((iv == NULL) ? 1 : !iv->initialized)
            return;

	if(show)
	{
	    /* Show toolbar. */
	    if(!iv->toolbar_map_state)
	    {
		GtkWidget *w = iv->toolbar_toplevel;
		if(w != NULL)
		    gtk_widget_show(w);
                iv->toolbar_map_state = TRUE;
                ImgViewUpdateMenus(iv);
	    }
        }
	else
	{
            /* Hide toolbar. */
            if(iv->toolbar_map_state)
            {
                GtkWidget *w = iv->toolbar_toplevel;
                if(w != NULL)
                    gtk_widget_hide(w);
                iv->toolbar_map_state = FALSE;
		ImgViewUpdateMenus(iv);
            }
	}
}


/*
 *	Maps the image viewer.
 */
void ImgViewMap(imgview_struct *iv)
{
        if((iv == NULL) ? 1 : !iv->initialized)
            return;

	if(!iv->map_state)
	{
	    GtkWidget *w = iv->toplevel;
	    if(w != NULL)
		gtk_widget_show(w);
	    iv->map_state = TRUE;
	}
}

/*
 *	Unmaps the image viewer.
 */
void ImgViewUnmap(imgview_struct *iv)
{
        if((iv == NULL) ? 1 : !iv->initialized)
            return;

        if(iv->map_state)
        {
            GtkWidget *w = iv->toplevel;
            if(w != NULL)
                gtk_widget_hide(w);
            iv->map_state = FALSE;
        }
}

/*
 *	Destroys the image viewer's resources and deallocates its
 *	structure.
 */
void ImgViewDelete(imgview_struct *iv)
{
	GdkFont **font;
	GdkGC **gc;
	GdkBitmap **bitmap;
	GdkPixmap **pixmap;
	GdkCursor **cur;
	GtkWidget **w;
	GtkAdjustment **adj;


        if(iv == NULL)
            return;

	if(iv->initialized)
	{
#define DO_UNREF_BITMAP		\
{ \
 if((*bitmap) != NULL) \
 { \
  GdkBitmap *tbm = *bitmap; \
  (*bitmap) = NULL; \
  gdk_bitmap_unref(tbm); \
 } \
}

#define DO_UNREF_PIXMAP		\
{ \
 if((*pixmap) != NULL) \
 { \
  GdkPixmap *tpm = *pixmap; \
  (*pixmap) = NULL; \
  gdk_pixmap_unref(tpm); \
 } \
}

#define DO_UNREF_FONT		\
{ \
 if((*font) != NULL) \
 { \
  GdkFont *tf = *font; \
  (*font) = NULL; \
  gdk_font_unref(tf); \
 } \
}

#define DO_UNREF_GC		\
{ \
 if((*gc) != NULL) \
 { \
  GdkGC *tgc = *gc; \
  (*gc) = NULL; \
  gdk_gc_unref(tgc); \
 } \
}

#define DO_DESTROY_CURSOR       \
{ \
 if((*cur) != NULL) \
 { \
  GdkCursor *tc = *cur; \
  (*cur) = NULL; \
  gdk_cursor_destroy(tc); \
 } \
}

#define DO_DESTROY_WIDGET       \
{ \
 if((*w) != NULL) \
 { \
  GtkWidget *tmp_w = *w; \
  (*w) = NULL; \
  gtk_widget_destroy(tmp_w); \
 } \
}

#define DO_UNREF_ADJUSTMENT	\
{ \
 if((*adj) != NULL) \
 { \
  GtkAdjustment *ta = *adj; \
  (*adj) = NULL; \
  gtk_object_unref(GTK_OBJECT(ta)); \
 } \
}
	    /* Reset the image viewer, this will unload the original image
	     * and deallocate other resource on the image viewer.
	     */
	    ImgViewReset(iv, iv->map_state);

	    /* Flush GDK before destroying widgets, incase any shared
	     * image puts are still happening.
	     */
	    gdk_flush();


	    /* Begin destroying widgets. */

	    w = &iv->menu;
	    iv->zoom_in_mi = NULL;
	    iv->zoom_out_mi = NULL;
	    iv->zoom_tofit_mi = NULL;
	    iv->zoom_onetoone_mi = NULL;
	    iv->show_toolbar_micheck = NULL;
	    iv->show_values_micheck = NULL;
	    DO_DESTROY_WIDGET

            w = &iv->toplevel;
            iv->main_vbox = NULL;
            iv->toolbar_toplevel = NULL;
            iv->info_label = NULL;
            iv->zoom_in_btn = NULL;
            iv->zoom_out_btn = NULL;
            iv->zoom_onetoone_btn = NULL;
            iv->zoom_tofit_btn = NULL;
            DO_DESTROY_WIDGET

	    pixmap = &iv->wm_icon_pixmap;
	    DO_UNREF_PIXMAP
	    bitmap = &iv->wm_icon_mask;
	    DO_UNREF_BITMAP

            if(iv->accelgrp != NULL)
            {
                gtk_accel_group_unref(iv->accelgrp);
                iv->accelgrp = NULL;
            }

	    if(iv->view_img != NULL)
	    {
		ImgViewImageDelete(iv->view_img);
		iv->view_img = NULL;
	    }

	    /* The orig_img should already be destroyed when we resetted. */

	    adj = &iv->view_x_adj;
	    DO_UNREF_ADJUSTMENT
            adj = &iv->view_y_adj;
            DO_UNREF_ADJUSTMENT

	    cur = &iv->translate_cur;
	    DO_DESTROY_CURSOR
	    cur = &iv->zoom_cur;
	    DO_DESTROY_CURSOR
	    cur = &iv->zoom_rectangle_cur;
	    DO_DESTROY_CURSOR

	    gc = &iv->view_selection_gc;
	    DO_UNREF_GC

	    font = &iv->view_font;
	    DO_UNREF_FONT

#undef DO_UNREF_BITMAP
#undef DO_UNREF_PIXMAP
#undef DO_UNREF_FONT
#undef DO_UNREF_GC
#undef DO_DESTROY_CURSOR
#undef DO_DESTROY_WIDGET
#undef DO_UNREF_ADJUSTMENT
	}

	/* Deallocate structure itself. */
	free(iv);
}
