#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#ifdef HAVE_IMLIB
# include <Imlib.h>
#endif
#ifdef HAVE_IMLIB2
# include <X11/Xlib.h>
# include <Imlib2.h>
#endif
#include <gtk/gtk.h>
#include <gdk/gdkx.h>

#include "guiutils.h"
#include "cdialog.h"
#include "fb.h"
#include "imgview.h"
#include "config.h"

#include "images/icon_open_20x20.xpm"


#ifdef HAVE_IMLIB
/* Handle used for passing to all Imlib functions, not needed for
 * Imlib2.
 */
void *imlib_handle = NULL;
#endif	/* HAVE_IMLIB */


/* Local global current directory. */
static gchar current_directory[PATH_MAX];


/* Define this to create the image viewer as a stand alone toplevel
 * window.
 */
#define CREATE_IMGVIEW_AS_TOPLEVEL	1



/*
 *	Called when the image viewer toplevel has closed.
 */
static gint delete_event(
	GtkWidget *widget, GdkEvent *event, gpointer data
)
{
        if(CDialogIsQuery())
            return(TRUE);

        if(FileBrowserIsQuery())
            return(TRUE);

	gtk_main_quit();
        return(TRUE);
}

/*
 *	Procedure to load an image to the given image viewer.
 */
static void do_load_image(
	imgview_struct *iv, const char *path
)
{
	GtkWidget *w;
	gint width, height;
#ifdef HAVE_IMLIB
	gchar *dpath;
	ImlibImage *imlib_image;
#endif
#ifdef HAVE_IMLIB2
	Imlib_Image imlib_image;
#endif


#ifdef HAVE_IMLIB
	if(imlib_handle == NULL)
	    return;
#endif

        if((iv == NULL) || (path == NULL))
	    return;

        /* Get image viewer toplevel widget if it is a GtkWindow. */
	if(ImgViewToplevelIsWindow(iv))
	    w = ImgViewGetToplevelWidget(iv);
	else
	    w = NULL;

#ifdef HAVE_IMLIB
	/* Copy path and load Imlib image. */
	dpath = g_strdup(path);
	imlib_image = Imlib_load_image(imlib_handle, dpath);
	g_free(dpath); dpath = NULL;
#endif
#ifdef HAVE_IMLIB2
	imlib_image = imlib_load_image(path);
#endif

	/* Imlib image loaded successfully? */
	if(imlib_image != NULL)
        {
#ifdef HAVE_IMLIB
	    /* Get size of newly loaded image. */
            width = imlib_image->rgb_width;
            height = imlib_image->rgb_height;
#endif
#ifdef HAVE_IMLIB2
	    /* Set newly loaded Imlib image into context. */
	    imlib_context_set_image(imlib_image);
	    /* Get size of newly loaded image. */
	    width = imlib_image_get_width();
	    height = imlib_image_get_height();
#endif

	    /* Image viewer toplevel window available? */
	    if(w != NULL)
	    {
		/* Image viewer's toplevel widget is a GtkWindow, now
		 * we know we can resize it. Get the view drawing area
		 * and set its size to match the new image's size.
		 */
		gchar *title;
		GtkWidget *w2 = (GtkWidget *)ImgViewGetViewWidget(iv);
		if(w2 != NULL)
		{
		    gtk_widget_set_usize(
                        w2, width, height
		    );
                    gtk_widget_queue_resize(w);	/* Notify toplevel to resize. */
		}

		/* Set title of image viewer. */
		title = (gchar *)g_malloc(strlen(path) + 256);
		if(title != NULL)
		{
		    const gchar *path_ptr = strrchr(path, '/');
		    if(path_ptr == NULL)
			path_ptr = path;
		    else
			path_ptr++;

		    sprintf(
			title,
			"Image Viewer: %s",
			path_ptr
		    );
		    gtk_window_set_title(GTK_WINDOW(w), title);

		    g_free(title);
		}
	    }

#ifdef HAVE_IMLIB
	    /* Load new image to image viewer, any currently loaded image
	     * will be unloaded first.
	     */
	    ImgViewLoad(
		iv,
		width, height,
		0,			/* Auto calculate bpl. */
		IMGVIEW_FORMAT_RGB,
		(const guint8 *)imlib_image->rgb_data
	    );
#endif
#ifdef HAVE_IMLIB2
            /* Load new image to image viewer, any currently loaded image
             * will be unloaded first.
             */
            ImgViewLoad(
                iv,
                width, height,
                0,                      /* Auto calculate bpl. */
                IMGVIEW_FORMAT_RGBA,
                (const guint8 *)imlib_image_get_data_for_reading_only()
	    );
#endif

#ifdef HAVE_IMLIB
	    /* Deallocate Imlib image, it is on longer needed. */
	    Imlib_destroy_image(imlib_handle, imlib_image);
#endif
#ifdef HAVE_IMLIB2
            /* Always deallocate Imlib2 images after rendering, this is
	     * efficient because the image is cached and managed by Imlib.
	     */
            imlib_free_image();
#endif
        }
	else
	{
	    /* Failed to load Imlib image. */
	    gint len = strlen(path) + (80 * 2);
	    gchar *buf;

	    buf = (gchar *)g_malloc(len);
	    if(buf != NULL)
		sprintf(
		    buf,
		    "Cannot open image:\n\n    %s",
		    path
		);
            CDialogSetTransientFor(w);
            CDialogGetResponse(
		"Open Failed",
		buf,
"Please check if the specified image file exists and\n\
is a valid format supported by Imlib.",
                CDIALOG_ICON_WARNING,
                CDIALOG_BTNFLAG_OK | CDIALOG_BTNFLAG_HELP,
                CDIALOG_BTNFLAG_OK
            );
            CDialogSetTransientFor(NULL);

	    g_free(buf);
	}
}

/*
 *	Load a new image callback.
 */
static void open_image(GtkWidget *widget, gpointer data)
{
	gchar **path_rtn; gint path_total_rtns;
	fb_type_struct *type_rtn;
	gchar *strptr;
	fb_type_struct **ftype = NULL; gint total_ftypes = 0;
	GtkWidget *w;
	gchar tmp_path[PATH_MAX + NAME_MAX];
	imgview_struct *iv = (imgview_struct *)data;
        if(iv == NULL)
            return;

	if(FileBrowserIsQuery())
	    return;

	/* Allocate file types list. */
	FileBrowserTypeListNew(
	    &ftype, &total_ftypes,
	    "*.*", "Images"
	);

	/* Get image viewer toplevel widget if it is a GtkWindow. */
        if(ImgViewToplevelIsWindow(iv))
            w = ImgViewGetToplevelWidget(iv);
        else
            w = NULL;

	/* Get new path from user. */
	FileBrowserSetTransientFor(w);
	if(FileBrowserGetResponse(
	    "Open Image", "Open", "Cancel",
	    current_directory,		/* Initial path. */
	    ftype, total_ftypes,
	    &path_rtn, &path_total_rtns,
	    &type_rtn
	))
	{
	    if(path_total_rtns > 0)
		strptr = path_rtn[0];
	    else
		strptr = NULL;
	    if(strptr != NULL)
	    {
		strncpy(tmp_path, strptr, PATH_MAX + NAME_MAX);
		tmp_path[PATH_MAX + NAME_MAX - 1] = '\0';


		/* Load image. */
		do_load_image(iv, tmp_path);


		/* Update current directory. */
		strncpy(current_directory, tmp_path, PATH_MAX);
		current_directory[PATH_MAX - 1] = '\0';

		strptr = strrchr(current_directory, '/');
		if(strptr != NULL)
		    (*strptr) = '\0';
		if((*current_directory) == '\0')
		    strcpy(current_directory, "/");
		current_directory[PATH_MAX - 1] = '\0';
	    }
	}
	FileBrowserSetTransientFor(NULL);

	/* Deallocate file types list. */
	FileBrowserDeleteTypeList(ftype, total_ftypes);
}

/*
 *	Parses the parameter and value from an argument, returning a
 *	statically allocated string or a string pointing within arg for
 *	both the parm_rtn and val_rtn.
 *
 *	The given arg will not be modified.
 */
static void get_args_value(
	char *arg, char **parm_rtn, char **val_rtn
)
{
	gchar *strptr;
	static gchar parm_str[256];


	if(parm_rtn != NULL)
	    (*parm_rtn) = "";
	if(val_rtn != NULL)
	    (*val_rtn) = "";

	if(arg == NULL)
	    return;

	/* If first char is not a '-' or '+' then skip as there is
	 * no parameter or value.
	 */
	if(((*arg) != '-') && ((*arg) != '+'))
	    return;

	/* Seek past initial '-' or '+' characters. */
	while(((*arg) == '-') || ((*arg) == '+'))
	    arg++;

	/* Get parameter and store it in our parm_str. */
	strncpy(parm_str, arg, 256);
	parm_str[256 - 1] = '\0';

	strptr = strchr(parm_str, '=');
	if(strptr != NULL)
	    (*strptr) = '\0';
	strptr = strchr(parm_str, '-');
        if(strptr != NULL)
            (*strptr) = '\0';

	/* Update parm return. */
	if(parm_rtn != NULL)
            (*parm_rtn) = parm_str;


	/* Get value. */
	strptr = strchr(parm_str, '=');
	if(strptr == NULL)
	    strptr = strchr(parm_str, '-');
	if(strptr == NULL)
	{
	    if(val_rtn != NULL)
                (*val_rtn) = "";
	}
	else
	{
            if(val_rtn != NULL)
                (*val_rtn) = strptr + 1;
	}
}

/*
 *	Returns TRUE if the given value is true, accepted values are
 *	"yes" "y" "on" "active" "enabled" or any non-zero value. Any
 *	other value will return FALSE.
 */
static gbool stristrue(const gchar *s)
{
	int c;


	if(s == NULL)
	    return(FALSE);

	c = (int)(*s);
	if((c == 'Y') || (c == 'y'))
	    return(TRUE);

	if(!strcasecmp(s, "on"))
	    return(TRUE);

	if(!strcasecmp(s, "active"))
            return(TRUE);

	if(!strcasecmp(s, "enabled"))
            return(TRUE);

	c = atoi(s);
	if(c != 0)
            return(TRUE);

	return(FALSE);
}

int main(int argc, char *argv[])
{
	gbool show_toolbar = TRUE;
	gbool show_values = TRUE;
	gbool show_image_on_wm_icon = FALSE;
	const gchar *startup_file = NULL;

	gint i;
	GtkWidget *w, *parent;
	GtkWidget *toplevel = NULL;
	GtkWidget *iv_toplevel, *menu;
	imgview_struct *iv;
	struct stat stat_buf;
	gint rtn_val = 0;


	/* Parse arguments? */
	for(i = 1; i < argc; i++)
	{
	    gchar *parm, *val;
	    gchar *arg_ptr = (gchar *)argv[i];
	    if(arg_ptr == NULL)
		continue;

	    get_args_value(arg_ptr, &parm, &val);

	    /* Help? */
	    if(!strcasecmp(parm, "help") ||
               !strcasecmp(parm, "h") ||
               !strcasecmp(parm, "?")
	    )
	    {
		printf(PROG_USAGE);
		return(rtn_val);
	    }
	    /* Version? */
	    else if(!strcasecmp(parm, "version"))
            {
		printf(
		    "%s Version %s\n%s",
		    PROG_NAME, PROG_VERSION,
		    PROG_COPYRIGHT
		);
		return(rtn_val);
            }
	    /* General enable. */
            else if(!strcasecmp(parm, "enable") ||
                    !strcasecmp(parm, "with")
	    )
	    {
                if(!strcasecmp(val, "show_image_on_wm_icon"))
                    show_image_on_wm_icon = TRUE;
                else if (!strcasecmp(val, "show_values"))
                    show_values = TRUE;
                else if (!strcasecmp(val, "show_toolbar"))
                    show_toolbar = TRUE;
            }
            /* General disable. */
            else if(!strcasecmp(parm, "disable") ||
                    !strcasecmp(parm, "without")
            )
            {
                if(!strcasecmp(val, "show_image_on_wm_icon"))
                    show_image_on_wm_icon = FALSE;
                else if (!strcasecmp(val, "show_values"))
                    show_values = FALSE;
                else if (!strcasecmp(val, "show_toolbar"))
                    show_toolbar = FALSE;
            }
            /* Show/hide toolbar? */
            else if(!strcasecmp(parm, "toolbar"))
            {
                if(stristrue(val))
                    show_toolbar = TRUE;
                else
                    show_toolbar = FALSE;
            }
            else if(!strcasecmp(parm, "t"))
            {
		show_toolbar = FALSE;
	    }
            /* Show/hide values? */
            else if(!strcasecmp(parm, "show_values"))
            {
                if(stristrue(val))
                    show_values = TRUE;
                else
                    show_values = FALSE;
            }
            else if(!strcasecmp(parm, "q"))
            {
                show_values = FALSE;
            }
            /* Show image on wm icon? */
            else if(!strcasecmp(parm, "imageonwmicon"))
            {
		if(stristrue(val))
		    show_image_on_wm_icon = TRUE;
		else
		    show_image_on_wm_icon = FALSE;
            }
            else if(!strcasecmp(parm, "i"))
            {
		show_image_on_wm_icon = TRUE;
	    }

	    /* File name (check if arg is not a toggle)? */
	    else if(((*arg_ptr) != '-') &&
                    ((*arg_ptr) != '+')
            )
            {
		startup_file = (const gchar *)arg_ptr;
            }
	    /* Unsupported option. */
	    else
	    {
		fprintf(stderr, "%s: Unsupported option.\n", parm);
		rtn_val = 1;
		return(rtn_val);
	    }
	}


	/* Get current directory. */
	if(getcwd(current_directory, PATH_MAX) == NULL)
	   strcpy(current_directory, "/");
	current_directory[PATH_MAX - 1] = '\0';

	/* Check if given file to open at startup exists. */
	if(startup_file != NULL)
	{
	    if(stat(startup_file, &stat_buf))
	    {
		fprintf(
		    stderr,
		    "%s: No such file.\n",
		    startup_file
		);
		rtn_val = 2;
		return(rtn_val);
	    }
	    if(S_ISDIR(stat_buf.st_mode))
	    {
                fprintf(
                    stderr,
                    "%s: Is a directory.\n",
                    startup_file
                );
                rtn_val = 2;
                return(rtn_val);
	    }
	}


	/* Initialize GTK+. */
	gtk_init(&argc, &argv);
	gdk_rgb_init();

#ifdef HAVE_IMLIB
	/* Initialize Imlib. */
	imlib_handle = Imlib_init(GDK_DISPLAY());
#endif	/*HAVE_IMLIB */

	/* Initialize our GUI utilities. */
	CDialogInit();
        FileBrowserInit();

#ifdef CREATE_IMGVIEW_AS_TOPLEVEL
	/* Create image viewer as a toplevel window. */
        iv = ImgViewNew(
            show_toolbar, 		/* Show toolbar. */
	    show_values,		/* Show values. */
	    show_image_on_wm_icon,	/* Show image on wm icon. */
	    TRUE,			/* Toplevel to be a GtkWindow? */
            &iv_toplevel
        );
        if(iv == NULL)
	{
	    rtn_val = 1;
            return(rtn_val);
	}

	/* Get image viewer's toplevel widget. */
	parent = w = toplevel = iv_toplevel;

	/* Connect "delete_event" signal for image viewer's toplevel. */
        gtk_signal_connect(
            GTK_OBJECT(w), "delete_event",
            GTK_SIGNAL_FUNC(delete_event), NULL
        );
#else
        /* Create image viewer as a child widget. */
	iv = ImgViewNew(
	    show_toolbar,		/* Show toolbar. */
	    show_values,		/* Show values. */
	    show_image_on_wm_icon,	/* Show image on wm icon. */
	    FALSE,			/* Toplevel to be a GtkWindow? */
	    &iv_toplevel
	);
	if(iv == NULL)
	{
	    rtn_val = 1;
	    return(rtn_val);
	}

	/* Create our toplevel window which will be the parent for
	 * the image viewer widget.
	 */
	gtk_widget_push_visual(gdk_rgb_get_visual());
	gtk_widget_push_colormap(gdk_rgb_get_cmap());
	toplevel = w = gtk_window_new(GTK_WINDOW_TOPLEVEL);
        gtk_widget_pop_visual();
        gtk_widget_pop_colormap();
	gtk_window_set_policy(
	    GTK_WINDOW(w), TRUE, TRUE, FALSE
	);
        gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(delete_event), NULL
	);
        gtk_container_border_width(GTK_CONTAINER(w), 0);
	parent = w;

	/* Place image viewer toplevel widget into our toplevel window. */
        gtk_container_add(GTK_CONTAINER(parent), iv_toplevel);

        /* Make image viewer toplevel widget eligable for mapping. */
        ImgViewMap(iv);
#endif


	/* Get menu from image viewer. */
	menu = (GtkWidget *)ImgViewGetMenuWidget(iv);
	if(menu != NULL)
	{
            GtkWidget *fw;
            gint accel_key;
            gpointer accel_group = NULL;
            guint accel_mods;
            u_int8_t **icon;
            const gchar *label = NULL;
            gpointer mclient_data = iv;
            void (*func_cb)(GtkWidget *w, gpointer) = NULL;

#define DO_ADD_MENU_ITEM_LABEL  \
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_LABEL, accel_group, \
  icon, label, accel_key, accel_mods, (void **)&fw, \
  mclient_data, func_cb \
 ); \
}
#define DO_ADD_MENU_SEP         \
{ \
 w = GUIMenuItemCreate( \
  menu, GUI_MENU_ITEM_TYPE_SEPARATOR, NULL, \
  NULL, NULL, 0, 0, NULL, \
  NULL, NULL \
 ); \
}

	    DO_ADD_MENU_SEP

	    icon = (u_int8_t **)icon_open_20x20_xpm;
	    label = "Open...";
	    accel_key = 0;
	    accel_mods = 0;
	    func_cb = open_image;
	    DO_ADD_MENU_ITEM_LABEL

#undef DO_ADD_MENU_ITEM_LABEL
#undef DO_ADD_MENU_SEP
	}


#ifdef CREATE_IMGVIEW_AS_TOPLEVEL
        /* Make image viewer toplevel widget eligable for mapping. */
        ImgViewMap(iv);
#else
	/* Map our toplevel window. */
        gtk_widget_show(toplevel);
#endif

        /* Load startup image if path is not NULL. */
        if(startup_file != NULL)
            do_load_image(iv, startup_file);


	/* Enter a GTK+ management loop. */
	gtk_main();


	/* Destroy image viewer widget. */
	ImgViewDelete(iv);

#ifndef CREATE_IMGVIEW_AS_TOPLEVEL
	/* Destroy our toplevel window. */
	gtk_widget_destroy(toplevel);
#endif

	/* Shutdown our GUI utilities. */
	CDialogShutdown();
	FileBrowserShutdown();

	/* GTK+ does not need to be shut down. */

#ifdef HAVE_IMLIB
	/* Imlib does not need to be shut down. */
	imlib_handle = NULL;
#endif	/* HAVE_IMLIB */

	return(rtn_val);
}
